//adac.c
//Initializes audio output, sets up timed calls of mixing function
//Bryan Topp <betopp@betopp.com> 2015

#include <stdio.h>
#include <string.h>
#include <sys/param.h>

unsigned char adac_testmode = 0; //Overrides currently-playing voices

//--------------------------------------------
//Table of currently-playing sounds to mix
//--------------------------------------------

#define MAX_VOICES 16
typedef struct voice_s
{
	const short *data; //Incremented as we play
	unsigned int remaining_samples; //Decremented as we play
	unsigned char vol_left;
	unsigned char vol_right;
} voice_t;
static voice_t active_voices[MAX_VOICES];

//Adds a new sound to the active voice table
void adac_play(const short *data, unsigned int length_in_samples, unsigned char vol_left, unsigned char vol_right)
{
	for(unsigned char v = 0; v < MAX_VOICES; v++)
	{
		//Put in the first available voice
		//Todo: reserved voices
		voice_t *vptr = &(active_voices[v]);
		if(vptr->remaining_samples == 0)
		{
			vptr->remaining_samples = length_in_samples;
			vptr->data = data;
			vptr->vol_left = vol_left;
			vptr->vol_right = vol_right;
			//printf("play! %d\n", length_in_samples);
			return;
		}
	}
	
	printf("Voice overflow!\n");
}

//============================================


//--------------------------------------------
//Actual buffers that the audio is played from
//--------------------------------------------

#define AUDIO_SLICE_SAMPLES 512
#define AUDIO_SLICE_BUFFERS 4
__attribute__((aligned(16),section(".l1.bss.B"))) static short audio_data[AUDIO_SLICE_BUFFERS][AUDIO_SLICE_SAMPLES][2]; //Ring buffer
static unsigned char next_slice_to_render; //The slice that we are about to write into
static unsigned char get_current_slice(); //Returns the slice currently being played

//Mixing function - reads the active voice table and composites a slice of audio
//TODO: Maybe rewrite this as Blackfin assembly? It's exactly what the chip is made for...
static void adac_render_slice(short *sliceout)
{
	for(unsigned short sample = 0; sample < AUDIO_SLICE_SAMPLES; sample++)
	{
		int left = 0;
		int right = 0;
		
		for(unsigned char voice = 0; voice < MAX_VOICES; voice++)
		{
			voice_t *vptr = &(active_voices[voice]);
			if(vptr->remaining_samples > 0)
			{
				left += ((int)(*vptr->data)) * vptr->vol_left;
				right += ((int)(*vptr->data)) * vptr->vol_right;
				vptr->data++;
				vptr->remaining_samples--;
				
				//if(vptr->remaining_samples == 0)
				//	printf("done\n");
			}
		}
		
		//Todo: Blackfin built-in saturating arithmetic
		left = MIN(left, 256 * 32767);
		right = MIN(right, 256 * 32767);
		
		left = MAX(left, 256 * -32768);
		right = MAX(right, 256 * -32768);
		
		sliceout[(sample*2)+0] = left / 256;
		sliceout[(sample*2)+1] = right / 256;
	}
}

//Called asynchronously (relative to actual playback), from the video line interrupt during vblank
//Determines if we need to render more, and does so, if necessary
void adac_pump()
{
	//Don't stomp the audio output - get_current_slice() reflects where the DMA currently reads
	if(next_slice_to_render == get_current_slice())
		return;
	
	short *slicebuffer = &(audio_data[next_slice_to_render][0][0]);
	
	//Test modes
	if(adac_testmode != 0)
	{
		//Render a test
		for(unsigned short sample = 0; sample < AUDIO_SLICE_SAMPLES; sample++)
		{			
			short tri = ((((((sample%16)<8)?(sample%16):(15-(sample%16))) - 4)) * 4096) + 2047;
			if(adac_testmode & 1)
				slicebuffer[sample*2] = tri;
			else
				slicebuffer[sample*2] = 0;
			
			if(adac_testmode & 2)
				slicebuffer[(sample*2) + 1] = tri;
			else
				slicebuffer[(sample*2) + 1] = 0;
			
		}
	}
	else
	{
		adac_render_slice(slicebuffer);
	}
	
	//printf("rendered slice %d\n", next_slice_to_render);
	//Next time, render the next one (chase the DMA transfer!)
	next_slice_to_render = (next_slice_to_render + 1) % 4;
}

//============================================


//--------------------------------------------
//Hardware/output setup
//--------------------------------------------

//We don't generate new audio by request of the audio output/streaming.
//Instead, we periodically check where the audio output is, and render more if we are close to underrun.
//(We need to make sure that audio rendering happens during vblank, to properly share CPU time.)

#ifndef JSYS_PC_MOCKUP //Real code first

	#include <sys/platform.h>

	void adac_init()
	{
		//Set up SPORT1 Primary TX for internally-clocked I2S at 24000Hz
	}

	static unsigned char get_current_slice()
	{
		return 0;
	}

#else //JSYS_PC_MOCKUP switch, mockup code follows

	#include <stdlib.h>
	#include <SDL2/SDL.h>	

	volatile unsigned char slice_sending_to_sdl = 0;
	void adac_sdlcb(void*  userdata, Uint8* stream, int len)
	{
		slice_sending_to_sdl = (slice_sending_to_sdl + 1) % 4;

		//printf("sending slice %d to sdl\n", slice_sending_to_sdl);
		memcpy(stream, &(audio_data[slice_sending_to_sdl][0][0]), AUDIO_SLICE_SAMPLES * 2 * sizeof(short));
	}

	static unsigned char get_current_slice()
	{
		return slice_sending_to_sdl;
	}

	void adac_init()
	{
		printf("Mockup: audio output through SDL2\n");
		SDL_Init(SDL_INIT_AUDIO);

		SDL_AudioSpec desired_spec, actual_spec;

		memset(&desired_spec, 0, sizeof(desired_spec));
		desired_spec.freq = 24000;
		desired_spec.format = AUDIO_S16LSB;
		desired_spec.channels = 2;
		desired_spec.samples = AUDIO_SLICE_SAMPLES * 2;
		desired_spec.callback = &adac_sdlcb;

		if(SDL_OpenAudio(&desired_spec, &actual_spec) < 0)
		{
			printf("Failed to open audio device. \"%s\"\n", SDL_GetError());
			exit(-1);
		}
		
		if(actual_spec.freq != desired_spec.freq || 
			actual_spec.channels != desired_spec.channels ||
			actual_spec.format != desired_spec.format )
		{
			printf("Failed to open S16LE 24KHz 2ch audio\n");
			exit(-1);
		}
		
		SDL_PauseAudio(0);
		
	}

#endif

//============================================
