//dram.c
//Functions for initializing the DRAM and loading its contents
//Bryan Topp <betopp@betopp.com> 2016

#include "dram.h"
#include "spr.h" //For textplane
#include "vdac.h"

#ifndef JSYS_PC_MOCKUP

//Real system code: needs to set up hardware
#include "spiutils.h"
#include <sys/platform.h>
#include <stdio.h>

void dram_init()
{	
	sprintf(spr_textplane[11] + 2, "SDRAM controller init...");
	printf("Setting up Blackfin SDRAM controller...\n");
	
	//Disable all asynchronous banks
	*pEBIU_AMGCTL = 0;
	
	//Clear SDRAM status register sticky bits
	*pEBIU_SDSTAT = 0xFFFF;
	
	//Refresh rate control [BF533 HW Ref p. 17-48]
	//F_SCLK = 98.304MHz
	//T_REF = 64ms
	//NRA = 8192
	//T_RAS = 5
	//T_RP = 3
	//RDIV = ((F_SCLK x T_REF ) / NRA) - (T_RAS + T_RP )
	*pEBIU_SDRRC = 760;
	
	//Set up bank - 4M x 16 x 4 banks - AS4C16M16SA
	*pEBIU_SDBCTL = EBE | EBSZ_32 | EBCAW_9; 
	
	//Timings for AS4C16M16SA-7 at 98.304MHz SCLK / power-on sequence
	*pEBIU_SDGCTL = SCTLE | CL_2 | TRAS_5 | TRP_3 | TRCD_3 | TWR_2 | PSS;
		
	printf("Triggering SDRAM power-on.\n");
	ssync();
	
	//Trigger a read from SDRAM to start it
	unsigned short a = *(volatile unsigned short*)0; //BREAKIN' THE LAW BREAKIN' THE LAW
	a = a; //Suppress compiler warning

	//Set up cache because DRAM GOTTA BE SO GODDAMN SLOW
	printf("Memory map setup commencing: turning off and clearing DCPLBs.\n");
	ssync();
	*pDMEM_CONTROL &= ~ENDCPLB;
	ssync();
	for(unsigned short i = 0; i < 16; i++)
	{
		printf("\tOld DCPLB %.2d = (A)%.8X | (D)%.8X\n", i, (unsigned int)(pDCPLB_ADDR0[i]), (unsigned int)(pDCPLB_DATA0[i]));
		pDCPLB_DATA0[i] = 0;
	}
	
	//No clue why this was breaking things on my board, for now. Just going to come back later.
	printf("Setting up new DCPLBs...\n");

	//4K pages covering Data A SRAM (0xFF800000 - 0xFF803FFF) (Excludes upper half which we use as cache!)
	*pDCPLB_DATA0 = 0x0001001F; *pDCPLB_ADDR0 = (unsigned int*)0xFF800000;
	*pDCPLB_DATA1 = 0x0001001F; *pDCPLB_ADDR1 = (unsigned int*)0xFF801000;
	*pDCPLB_DATA2 = 0x0001001F; *pDCPLB_ADDR2 = (unsigned int*)0xFF802000;
	*pDCPLB_DATA3 = 0x0001001F; *pDCPLB_ADDR3 = (unsigned int*)0xFF803000;
	
	//4K pages covering Data B SRAM (0xFF900000 - 0xFF903FFF) (Excludes upper half which we use as cache!)
	*pDCPLB_DATA4 = 0x0001001F; *pDCPLB_ADDR4 = (unsigned int*)0xFF900000;
	*pDCPLB_DATA5 = 0x0001001F; *pDCPLB_ADDR5 = (unsigned int*)0xFF901000;
	*pDCPLB_DATA6 = 0x0001001F; *pDCPLB_ADDR6 = (unsigned int*)0xFF902000;
	*pDCPLB_DATA7 = 0x0001001F; *pDCPLB_ADDR7 = (unsigned int*)0xFF903000;
	
	//4M pages covering DRAMFS - cacheable!
	*pDCPLB_DATA8 = 0x0003109F; *pDCPLB_ADDR8 = (unsigned int*)0x00000000;
	*pDCPLB_DATA9 = 0x0003109F; *pDCPLB_ADDR9 = (unsigned int*)0x00400000;
	*pDCPLB_DATA10 = 0x0003109F; *pDCPLB_ADDR10 = (unsigned int*)0x00800000;
	*pDCPLB_DATA11 = 0x0003109F; *pDCPLB_ADDR11 = (unsigned int*)0x00C00000;
	*pDCPLB_DATA12 = 0x0003109F; *pDCPLB_ADDR12 = (unsigned int*)0x01000000;
	*pDCPLB_DATA13 = 0x0003109F; *pDCPLB_ADDR13 = (unsigned int*)0x01400000;
	
	//MMRs
	*pDCPLB_DATA14 = 0x0003001F; *pDCPLB_ADDR14 = (unsigned int*)0xFFC00000;
	
	//Scratchpad
	*pDCPLB_DATA15 = 0x0001001F; *pDCPLB_ADDR15 = (unsigned int*)0xFFB00000;
	
	for(unsigned short i = 0; i < 16; i++)
	{
		printf("\tNew DCPLB %.2d = (A)%.8X | (D)%.8X\n", i, (unsigned int)(pDCPLB_ADDR0[i]), (unsigned int)(pDCPLB_DATA0[i]));
	}	
	
	
	printf("DCPLB on...\n");
	ssync();
	*pDMEM_CONTROL |= ENDCPLB;
	ssync();
	
	printf("Cache on...\n");
	*pDMEM_CONTROL |= (1<<DMC0_P) | (1<<DMC1_P);
	*pDMEM_CONTROL &= ~DCBS;
	
	
	sprintf(spr_textplane[11] + 27, "OK");
	printf("Memory setup successful.\n");
	
	sprintf(spr_textplane[12] + 2, "Data load...");
	
	
	printf("Grabbing Flash ROM ID.\n");
	
	spi_set_rom();
	spi_exchange(0x90); //REMS
	spi_exchange(0x00);
	spi_exchange(0x00);
	spi_exchange(0x00);
	unsigned char rems_mfg = spi_exchange(0x00);
	unsigned char rems_dev = spi_exchange(0x00);
	spi_set_none();
	
	printf("REMS: MFG=0x%.2X DEV=0x%.2X\n", rems_mfg, rems_dev);
	
	if(rems_mfg != 0x01 || rems_dev != 0x17)
	{
		sprintf(spr_textplane[12] + 14, "REMS FAIL! Check ROM.");
		printf("Bad REMS ID... counterfeit?\n");
		return;
	}
	
	spi_set_rom();
	spi_exchange(0x9F); //RDID
	unsigned char rdid_id_0 = spi_exchange(0x00);
	unsigned char rdid_id_1 = spi_exchange(0x00);
	unsigned char rdid_id_2 = spi_exchange(0x00);
	unsigned char rdid_id_3 = spi_exchange(0x00);
	spi_set_none();
	
	unsigned int rdid_id_all = rdid_id_0 | (rdid_id_1<<8) | (rdid_id_2<<16) | (rdid_id_3<<24);
	
	printf("RDID: ID=0x%.8X\n", rdid_id_all);
	
	printf("Reading contents of flash ROM.\n");
	
	for(unsigned int megabyte = 0; megabyte < 16; megabyte++)
	{
		printf("Megabyte %u...\n", megabyte);
		spi_set_rom();
		
		//Read a megabyte at a time, starting 128K into Flash (beyond the BF533 LDR)
		unsigned int srcaddr = (megabyte * 1048576) + 131072;
		unsigned char *dstptr = DRAMFS_LOC + (megabyte * 1048576);
		
		//Send the read command and address
		spi_exchange(0x13); //4READ
		spi_exchange(srcaddr >> 24);
		spi_exchange(srcaddr >> 16);
		spi_exchange(srcaddr >> 8);
		spi_exchange(srcaddr >> 0);
		
		//Do this the dumb way for now.
		for(unsigned int c = 0; c < 1048576; c++)
		{
			*(dstptr++) = spi_exchange(0x00);
		}
		
		spi_set_none();
		sprintf(spr_textplane[12] + 16 + megabyte, "=");
		sprintf(spr_textplane[12] + 14, "%c", "/-\\|"[megabyte % 4]);
	}
	
	sprintf(spr_textplane[12] + 14, "Done!                    ");
	
	printf("Done.\n");
}

#else //Mockup code follows

#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
void *mockup_dram = NULL;

void dram_init()
{	
	sprintf(spr_textplane[11] + 2, "PC mockup loading DRAM blob.");

	//This is much simpler than on real hardware, don't you think?
	mockup_dram = malloc(DRAM_SIZE);
	if(!mockup_dram)
	{
		sprintf(spr_textplane[12] + 2, "Failed malloc.");
		printf("Mockup: failed to allocate memory for fake DRAM.\n");
		return;
	}

	//In the mockup, we load the DRAMFS blob from the filesystem.
	printf("Mockup: loading DRAMFS image from dramfs/dramfs.bin.\n");
	FILE *inf = fopen("dramfs/dramfs.bin", "rb");
	if(!inf)
	{
		sprintf(spr_textplane[12] + 2, "Failed fopen.");
		printf("Mockup: failed to open DRAMFS blob.\n");
		return;
	}
	
	fseek(inf, 0, SEEK_END);
	size_t filesize = ftell(inf);
	fseek(inf, 0, SEEK_SET);
	
	size_t bytes_read = fread(DRAMFS_LOC, 1, filesize, inf);
	if(bytes_read != filesize)
	{
		sprintf(spr_textplane[12] + 2, "Failed fread.");
		printf("Mockup: could not read whole DRAMFS blob.\n");
		return;
	}
	
	sprintf(spr_textplane[12] + 2, "Fake DRAM ready.");
	printf("Mockup: read %lu bytes of DRAMFS blob into fake DRAM.\n", bytes_read);
	fclose(inf);
	
}

#endif //JSYS_PC_MOCKUP