//build_dramfs.c
//Builds the DRAM-FS blob
//Bryan Topp <betopp@betopp.com> 2015

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <assert.h>


typedef struct enumerated_file_s
{
	const char *realname;
	const char *virtname;
	size_t length;
	struct enumerated_file_s *next;
} enumerated_file;

enumerated_file *enumerated_files = NULL;


void AddDir(const char *realdir, const char *virtdir)
{
	printf("Adding directory \"%s\" at \"%s\"\n", realdir, virtdir);
	
	//Iterate over files in the real directory
	DIR *dirf = opendir(realdir);
	assert(dirf);
	
	struct dirent *dp;
	while( (dp = readdir(dirf)) != NULL)
	{
		if(dp->d_name[0] == '.')
			continue;
				
		//Build full real and virtual names of file/directory
		char *full_real_name = (char*)malloc(strlen(realdir) + strlen(dp->d_name) + 3);
		snprintf(full_real_name, strlen(realdir) + strlen(dp->d_name) + 2, "%s/%s", realdir, dp->d_name);
		
		char *full_virt_name = (char*)malloc(strlen(virtdir) + strlen(dp->d_name) + 3);
		snprintf(full_virt_name, strlen(virtdir) + strlen(dp->d_name) + 2, "%s/%s", virtdir, dp->d_name);
		
		//Stat it to see if it's a file or directory
		struct stat statbuf;
		if(stat(full_real_name, &statbuf) < 0)
		{
			printf("failed to stat %s\n", full_real_name);
			assert(0);
		}
		
		if((statbuf.st_mode & S_IFMT) == S_IFDIR)
		{
			//Recurse
			AddDir(full_real_name, full_virt_name);
			//We don't add directories, so we're done with these names for now
			free(full_real_name);
			free(full_virt_name);
		}
		else
		{
			//Add the file to our list
			enumerated_file *newf = (enumerated_file*)malloc(sizeof(enumerated_file));
			assert(newf != NULL);
			newf->realname = full_real_name;
			newf->virtname = full_virt_name;
			newf->length = statbuf.st_size;
			newf->next = NULL;
			
			if(enumerated_files == NULL)
			{
				enumerated_files = newf;
			}
			else
			{
				enumerated_file *f = enumerated_files;
				while(f->next != NULL)
					f = f->next;
				
				f->next = newf;
			}
		}
	}
	
	closedir(dirf);
}

char zeroes[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

int main(int argc, const char **argv)
{
	if(argc != 3)
	{
		printf("Usage: %s <root dir> <output>\n", argv[0]);
		return -1;
	}
	
	const char *rootdir = argv[1];
	const char *outputname = argv[2];
	
	//Collect names and lengths of files
	printf("Beginning traversal...\n");
	
	AddDir(rootdir, "");
	
	/*printf("Collected files:\n");
	enumerated_file *f = enumerated_files;
	while(f != NULL)
	{
		printf("%s added as %s\n", f->realname, f->virtname);
		f = f->next;
	}*/
	
	//Selection sort
	printf("Sorting files...\n");
	
	enumerated_file *target = enumerated_files;
	while(target != NULL)
	{
		//Find the earliest name, at or past the target
		enumerated_file *search = target;
		enumerated_file *best = target;
		while(search != NULL)
		{
			if(strcmp(best->virtname, search->virtname) > 0)
				best = search;
			
			search = search->next;
		}
		
		//Swap it to the target
		const char *temp;
		temp = target->realname;
		target->realname = best->realname;
		best->realname = temp;
		temp = target->virtname;
		target->virtname = best->virtname;
		best->virtname = temp;
		size_t templ = target->length;
		target->length = best->length;
		best->length = templ;
		
		//Advance the target
		target = target->next;
		
		//(You know how a selection sort works, right?)
	}
	
	//Chop the initial slashes and count how many good files we have.
	//Total up the length required to store the filenames.
	printf("Removing leading slashes...\n");
	
	enumerated_file *f = enumerated_files;
	unsigned int counted_files = 0;
	size_t total_filename_length = 0;
	
	while(f != NULL)
	{
		assert(f->virtname[0] == '/');
		f->virtname++;
		
		total_filename_length += strlen(f->virtname) + 1;
		
		counted_files++;
		f = f->next;
	}
	
	/*printf("Sorted files:\n");
	f = enumerated_files;
	while(f != NULL)
	{
		printf("%s (%s)\n", f->virtname, f->realname);
		f = f->next;
	}*/
	
	//Open output and write magic
	printf("Opening output file...\n");
	FILE *outf = fopen(outputname, "wb");
	assert(outf);
	
	const char *magic = "betdramfs_00";
	fwrite(magic, 1, 12, outf);
	
	//Output count of files
	fwrite(&counted_files, 1, sizeof(unsigned int), outf);
	
	//Output directory, based on running size of filenames + file contents
	printf("Writing %d filenames...\n", counted_files);
	
	//Store intended positions of filenames and contents based on length
	f = enumerated_files;
	int i = 0;
	size_t running_count_filenames = 12 + sizeof(unsigned int) + (sizeof(unsigned int)*counted_files*3); //This is where filenames will start	
	size_t running_count_contents = running_count_filenames + total_filename_length; //This is where the contents starts
	if(running_count_contents % 16 != 0) //We align to 16 bytes before outputting each file's contents
		running_count_contents += 16 - (running_count_contents % 16);
	
	while(f != NULL)
	{
		assert(i < counted_files);
		
		//Output the running counts for each file in the DRAMFS
		fwrite(&running_count_filenames, 1, sizeof(unsigned int), outf);
		fwrite(&running_count_contents, 1, sizeof(unsigned int), outf);
		
		unsigned int length_to_write = f->length;
		fwrite(&length_to_write, 1, sizeof(unsigned int), outf);
				
		running_count_filenames += strlen(f->virtname) + 1;
		
		unsigned int padded_length = f->length;
		if(padded_length % 16 != 0) //We align to 16 bytes before outputting each file's contents
			padded_length += 16 - (padded_length % 16);
		
		running_count_contents += padded_length;
		
		i++;
		f = f->next;
	}

	//Then output the filenames themselves
	f = enumerated_files;
	while(f != NULL)
	{
		fwrite(f->virtname, 1, strlen(f->virtname)+1, outf);
		
		f = f->next;
	}		
	
	//Then output the file contents themselves
	f = enumerated_files;
	while(f != NULL)
	{
		if(ftell(outf) % 16 != 0)
		{
			fwrite(zeroes, 1, 16 - (ftell(outf)%16), outf);
		}
		
		
		FILE *inpf = fopen(f->realname, "rb");
		assert(inpf);
		
		unsigned char tempbuf[1024];
		size_t bytes_copied = 0;
		size_t bytes_read;
		
		while(bytes_copied < f->length)
		{
			bytes_read = fread(tempbuf, 1, 1024, inpf);
			assert(bytes_read >= 0);
			bytes_copied += fwrite(tempbuf, 1, bytes_read, outf);
		}
				
		fclose(inpf);
		
		f = f->next;
	}
	
	fclose(outf);
	return 0;
}