//fram.c
//FRAM interface - nonvolatile configuration/scores
//Bryan Topp <betopp@betopp.com> 2016

#include "fram.h"
#include "spr.h" //For textplane
#include <string.h>
#include <stdio.h>

//FRAM init and checksum code is same across real hardware / mockup

fram_t fram_shadow;

unsigned int fram_getchecksum()
{
	//Dumb checksum
	unsigned int *checksum_addr = (unsigned int*)(&fram_shadow);
	unsigned int subchecksum = 'Topp';
	unsigned int checksum = '1337';
	for(unsigned short b = 0; b < sizeof(fram_shadow) / sizeof(checksum); b++)
	{
		subchecksum += checksum_addr[b];
		checksum += subchecksum;
	}	
	
	return checksum;
}

void fram_putchecksum()
{	
	if(fram_getchecksum() == 0)
	{
		printf("FRAM checksum unchanged.\n");
	}
	else
	{
		//Depends on the checksum being the very last value!
		fram_shadow.checksum -= fram_getchecksum();
		printf("FRAM checksum updated.\n");
	}
}


static const fram_t fram_initial = {
	.highscores = {	
		{.score = 10000, .name = "---", .maxlevel = 1 },
		{.score = 9000,  .name = "---", .maxlevel = 1 },
		{.score = 8000,  .name = "---", .maxlevel = 1 },
		{.score = 7000,  .name = "---", .maxlevel = 1 },
		{.score = 6000,  .name = "---", .maxlevel = 1 },
		{.score = 5000,  .name = "---", .maxlevel = 1 },
		{.score = 4000,  .name = "---", .maxlevel = 1 },
		{.score = 3000,  .name = "---", .maxlevel = 1 },
		{.score = 2000,  .name = "---", .maxlevel = 1 },
		{.score = 1000,  .name = "---", .maxlevel = 1 } }
};

void fram_init()
{
	
	sprintf(spr_textplane[8]+2, "FRAM init.");
	printf("FRAM setup. Read old contents into shadow...\n");
	
	fram_read();
	
	//Check that the FRAM is valid.	
	if(fram_getchecksum() != 0)
	{
		sprintf(spr_textplane[9]+2, "Content invalid; cleared.");
		fprintf(stderr, "FRAM checksum invalid. Initializing...\n");
		memcpy(&fram_shadow, &fram_initial, sizeof(fram_initial));	
	}
	else
	{
		sprintf(spr_textplane[9]+2, "Content OK.");
		printf("FRAM checksum valid on init.\n");
	}
	
	fram_write();
}

void fram_nuke()
{
	printf("Force re-init of FRAM.\n");
	
	printf("In-memory clear...\n");
	memcpy(&fram_shadow, &fram_initial, sizeof(fram_initial));

	fram_write();
}

void fram_insert_highscore(unsigned int score, unsigned char *name, unsigned char maxlevel)
{
	//Search from the top of the board down, to find a score we're ahead of
	unsigned char newpos;
	for(newpos = 0; newpos < FRAM_NUM_SCORES; newpos++)
	{
		if(score > fram_shadow.highscores[newpos].score)
			break;
	}
	
	if(newpos >= FRAM_NUM_SCORES) //Off the bottom of the board, too bad.
		return;
	
	//Bump all the scores below the new one
	//(From the bottom up, overwrite each with the one above.)
	for(unsigned char bumpto = FRAM_NUM_SCORES-1; bumpto > newpos; bumpto--)
	{
		fram_shadow.highscores[bumpto] = fram_shadow.highscores[bumpto-1];
	}
	
	//Score
	fram_shadow.highscores[newpos].score = score;
	
	//Name, and enforce a NUL terminator
	memcpy(fram_shadow.highscores[newpos].name, name, 3);
	fram_shadow.highscores[newpos].name[3] = '\0';
	
	//Max level
	fram_shadow.highscores[newpos].maxlevel = maxlevel;	
}


#ifndef JSYS_PC_MOCKUP

//Actual hardware - read contents of FRAM chip over SPI
#include "spiutils.h"

void fram_read()
{	
	printf("FRAM read start.\n");
		
	spi_set_fram();
	printf("Getting FRAM ID...\n");
	
	spi_exchange(0x9F); //RDID
	unsigned char rdid_id_0 = spi_exchange(0x00);
	unsigned char rdid_id_1 = spi_exchange(0x00);
	unsigned char rdid_id_2 = spi_exchange(0x00);
	unsigned char rdid_id_3 = spi_exchange(0x00);	
	unsigned int rdid_id_all = rdid_id_0 | (rdid_id_1<<8) | (rdid_id_2<<16) | (rdid_id_3<<24);
	
	spi_set_none();
	printf("RDID: ID=0x%.8X\n", rdid_id_all);
	
	spi_set_fram();
	printf("Getting status register...\n");
	
	spi_exchange(0x05); //RDSR
	unsigned char statusreg = spi_exchange(0x00);	
	
	spi_set_none();
	printf("RDSR: SR=0x%.2X\n", statusreg);
	
	printf("Reading %lu bytes...\n", sizeof(fram_shadow));
	spi_set_fram();
	
	//Send command
	spi_exchange(0x03); //READ
	spi_exchange(0x00); //High address bits
	spi_exchange(0x00); //Low address bits
	
	//Read bytes
	for(unsigned short byte = 0; byte < sizeof(fram_shadow); byte++)
	{
		((char*)(&fram_shadow))[byte] = spi_exchange(0x00);
	}
	
	printf("Done.\n");
	
	spi_set_none();
}

void fram_write()
{
	fram_putchecksum();
	printf("FRAM write start.\n");
	
	spi_set_fram();
	printf("Setting write-enable latch...\n");
	
	spi_exchange(0x06); //WREN
	//No parameters.
	
	spi_set_none();
	
	printf("Writing %lu bytes...\n", sizeof(fram_shadow));
	spi_set_fram();
	
	//Send command
	spi_exchange(0x02); //WRITE
	spi_exchange(0x00); //High address bits
	spi_exchange(0x00); //Low address bits
	
	//Write bytes
	for(unsigned short byte = 0; byte < sizeof(fram_shadow); byte++)
	{
		spi_exchange(((char*)(&fram_shadow))[byte]);
	}
	
	printf("Done.\n");

	spi_set_none();
}

#else //Mockup code follows

#include <stdio.h>

void fram_read()
{
	printf("Mockup: reading fram.bin.\n");
	FILE *frambin = fopen("fram.bin", "rb");
	if(frambin == NULL)
	{
		printf("Mockup: failed to open fram.bin for reading.\n");
	}
	else
	{
		size_t bytes_read = fread(&fram_shadow, 1, sizeof(fram_shadow), frambin);
		if(bytes_read != sizeof(fram_shadow))
			printf("Mockup: failed to read from fram.bin.\n");
	
		fclose(frambin);
	}
}

void fram_write()
{
	printf("FRAM write start.\n");
	fram_putchecksum();

	printf("Mockup: writing fram.bin.\n");
	FILE *frambin = fopen("fram.bin", "wb");
	if(frambin == NULL)
	{
		printf("Mockup: failed to open fram.bin for writing.\n");
	}
	else
	{
		size_t bytes_written = fwrite(&fram_shadow, 1, sizeof(fram_shadow), frambin);
		if(bytes_written != sizeof(fram_shadow))
			printf("Mockup: failed to write to fram.bin.\n");
	
		fclose(frambin);
	}
}

#endif //JSYS_PC_MOCKUP switch