//main.c
//Entry points for Blackfin video game
//Bryan Topp <betopp@betopp.com> 2015

#include <stdio.h>

#ifdef JSYS_PC_MOCKUP
	#include <SDL2/SDL.h>
#else
	#include <sys/platform.h>
#endif

//Primary systems: I/O drivers
#include "uart.h"
#include "vdac.h"
#include "pads.h"
#include "dram.h"
#include "fram.h"
#include "adac.h"

//Secondary systems: software support
#include "spr.h"
#include "dramfs.h"

//Application tasks
#include "testmenu.h" //Test menu
#include "padloader.h" //IO controller reflasher
extern void shooty(); //Video game
extern void exception_init(); //Exception handler


//Startup entry point.
int main()
{
	
#ifndef JSYS_PC_MOCKUP
	//Initialize the CPU clocks
	
	//CLKIN on the BF533GS board is 12.288MHz
	//Set VCO to CLKIN * 32 = 393MHz
	*pPLL_CTL = SET_MSEL(32); //See ADSP-BF533 HW Ref p. 8-8
	
	//PLL programming sequence, see ADSP-BF533 HW Ref p. 8-21
	asm("CLI R0 ;");
	asm("IDLE ;");
	asm("STI R0 ;");
	
	//Set CCLK to VCO/1 = 393MHz
	//Set SCLK to VCO/4 = 98MHz
	*pPLL_DIV = SET_SSEL(4); //See ADSP-BF533 HW Ref p. 8-7
	
	//(PPI is clocked externally by the CLKIN through a divide-by-2 (DFF) = 6.144MHz)
#endif
	
	//Initialize UART and set up exception handler for the rest of the process
	uart_init();
	exception_init();
	
	//Set up DAC DMA chains
	vdac_init();
	adac_init();
	
	//Initialize rendering system - following systems may want to show text
	spr_init(); //(Writes things on textplane lines 2 and 3)
	
	//Show version info splash
	sprintf(spr_textplane[24]+2, "BF533GS ROM");
	sprintf(spr_textplane[25]+2, JSYS_GIT_VERSION);
	sprintf(spr_textplane[26]+2, JSYS_BUILD_DATE);
	sprintf(spr_textplane[27]+2, JSYS_BUILDER);
	
	pads_init(); //inputs / coinage (Writes things on textplane lines 5 and 6)
	fram_init(); //configuration data (Writes things on textplane lines 8 and 9)
	dram_init(); //memory controller (Writes things on textplane lines 11 and 12)
	dramfs_init(); //memory-resident filesystem (Writes things on textplane lines 14 and 15)

	//Wait a bit for the init text to sink in
	printf("Waiting a bit for the init text to be readable.\n");
	for(int i = 0; i < 120; i++)
		vdac_fieldwait();
	printf("Waited sufficiently for video sync, continuing.\n");
	
	if(!pads_isvalid())
	{
		//If the IO controller failed, complain about that
		fprintf(stderr, "IOCON not valid, starting padloader.\n");
		padloader();
	}
	else if(!dramfs_isvalid())
	{
		//If the DRAMFS isn't valid, we can still at least run the test menu
		fprintf(stderr, "DRAMFS not valid, force testmenu.\n");
		testmenu();
	}
	else
	{
		//If init went OK, then enter the game or test menu depending on the state of the test switch
		printf("Reading state of test switch.\n");
		vdac_fieldwait();
		pads_poll();
		vdac_fieldwait();
		pads_poll();
		if( pads_sysbuttons & PADS_SYS_BIT_TEST )
		{
			printf("Test switch ON. Executing test menu.\n");
			testmenu();
		}
		else
		{
			printf("Test switch OFF. Executing game.\n");
			shooty();
		}
	}
	
	//Something bad happened: reset
#ifdef JSYS_PC_MOCKUP
	//Just fall through and quit the mockup
	printf("main() fallthrough: mockup quit.\n");
#else
	//Soft-reset the system on real hardware
	printf("main() fallthrough: soft reset.\n");
	
	*pSWRST = 7;
	ssync();
	for(short i = 0; i < 100; i++)
		asm("nop;");
		
	*pSWRST = 0;
	ssync();
	asm("raise 1;");
#endif
	return 0;
}