//padloader.c
//Reflash code for the IO controller
//Bryan Topp <betopp@betopp.com> 2016

#include "padloader.h"
#include "spr.h"

#ifndef JSYS_PC_MOCKUP //Blackfin code on real system

#include <sys/platform.h>
#include <string.h>
#include <stdio.h>
#include "spiutils.h"

//#define IOCON_EMBEDDED

#ifdef IOCON_EMBEDDED
	__attribute__((aligned(16),section(".l1.data.B")))
	#include "../iotiny/iocon.array"
	static unsigned short iocon_embedded_count = 0;
#endif //IOCON_EMBEDDED

void padloader()
{
	//Buffer to hold base64 data, allowing UART to feed IOCON programming data
	unsigned char uart_base64_decoded[1024 + 4] = {0}; //Leave space at end for decoding long data
	unsigned short uart_base64_decoded_pos = 0;

	unsigned char uart_base64_incoming[4] = {0};
	unsigned short uart_base64_incoming_pos = 0;

	static const char * const message[6] =
	{
		"+------[FAILURE DETECTED]------+",
		"|                              |",
		"|      IOCON unresponsive.     |",
		"|    Board requires service.   |",
		"|                              |",
		"+------------------------------+"
	};
	
	spr_textplane_clear();
	spr_sprite_clear();
	spr_background = NULL;
	
	for(unsigned char m = 0; m < 6; m++)
	{
		memcpy(spr_textplane[6+m] + 4, message[m], 32);
	}

	fprintf(stderr, "IOCON needs reflash.\nPlease provide 1KB image via Base64 data on UART.\n");
	
	//Turn off the divisor-latch access byte (let us read the serial data)
	*pUART_LCR &= ~DLAB;
	
	while(uart_base64_decoded_pos < 1024)
	{		
		#ifndef IOCON_EMBEDDED
			//If there's nothing new to read (DR flag clear), don't repeatedly read the same byte
			while(!(*pUART_LSR & DR))
			{	
				//Spin
			}
			
			//Grab new byte, clearing DR flag
			unsigned char newbyte = *pUART_RBR;
			
		#else //IOCON_EMBEDDED follows
			unsigned char newbyte = iocon_base64[iocon_embedded_count++];
		#endif
		
		//Store it in our incoming buffer if it's a base64 character - reset otherwise
		if(newbyte >= 'A' && newbyte <= 'Z')
			uart_base64_incoming[uart_base64_incoming_pos++] = 0 + newbyte - 'A';
		else if(newbyte >= 'a' && newbyte <= 'z')
			uart_base64_incoming[uart_base64_incoming_pos++] = 26 + newbyte - 'a';
		else if(newbyte >= '0' && newbyte <= '9')
			uart_base64_incoming[uart_base64_incoming_pos++] = 52 + newbyte - '0';
		else if(newbyte == '+')
			uart_base64_incoming[uart_base64_incoming_pos++] = 62;
		else if(newbyte == '/')
			uart_base64_incoming[uart_base64_incoming_pos++] = 63;
		else if(newbyte == '=')
			uart_base64_incoming[uart_base64_incoming_pos++] = 64;
		else
		{
			uart_base64_incoming[0] = 0;
			uart_base64_incoming[1] = 0;
			uart_base64_incoming[2] = 0;
			uart_base64_incoming[3] = 0;
			
			uart_base64_incoming_pos = 0;			
			uart_base64_decoded_pos = 0;
		}
		
		//If we've got 4 base64 characters received, decode
		if(uart_base64_incoming_pos == 4)
		{
			uart_base64_incoming_pos = 0; //Next 4 characters start anew
			
			unsigned char a = uart_base64_incoming[0];
			unsigned char b = uart_base64_incoming[1];
			unsigned char c = uart_base64_incoming[2];
			unsigned char d = uart_base64_incoming[3];
			
			uart_base64_decoded[uart_base64_decoded_pos++] = (a << 2) | (b >> 4);
			if (c < 64)
			{
				uart_base64_decoded[uart_base64_decoded_pos++] = (b << 4) | (c >> 2);
				if (d < 64)
				{
					uart_base64_decoded[uart_base64_decoded_pos++] = (c << 6) | d;
				}
			}
		}
	}
	
	printf("Image received. Thank you.\n");
	
	//SPI setup: not using any SPI chip-selects right now - IOCON doesn't care
	*pSPI_FLG = 0xFF00;
	*pSPI_BAUD = 256; //Go real slow
	*pSPI_CTL = MSTR | SPE | SZ | TDBR_CORE; //CPHA=CPOL=0
	
	*pFIO_DIR = ~0; //All flags output	

	*pFIO_FLAG_D = 0xFFFF; //All flags high
	ssync();
	for(int nop = 0; nop < 3500000; nop++){	asm("nop;"); }
	
	*pFIO_FLAG_D = 0xFFFE; //PF0 low, putting IOCON in reset
	ssync();
	for(int nop = 0; nop < 3500000; nop++){	asm("nop;"); }

	//Send programming sequence
	unsigned char a = spi_exchange(0xAC);	
	unsigned char b = spi_exchange(0x53);
	unsigned char c = spi_exchange(0x00);
	unsigned char d = spi_exchange(0x00);
	
	printf("IOCON ident: 0x%.2X 0x%.2X 0x%.2X 0x%.2X\n", a, b, c, d);
	
	if(c != 0x53)
	{
		fprintf(stderr, "Error. IOCON did not echo programming sequence.\n");
		return;
	}
	
	printf("Sending chip-erase command.\n");
	
	spi_exchange(0xAC);
	spi_exchange(0x80);
	spi_exchange(0x00);
	spi_exchange(0x00);
	
	printf("Waiting for chip to be ready.\n");
	
	while(1)
	{
		spi_exchange(0xF0);
		spi_exchange(0x00);
		spi_exchange(0x00);
		unsigned char rdybsy = spi_exchange(0x00);
		
		if( (rdybsy & 0x01) == 0 ) //LSB is set to 0 when done
			break;
	}
	
	printf("Chip indicates ready. Starting program.\n");
	
	//64-byte pages
	for(unsigned char page = 0; page < 16; page++)
	{
		printf("Page %d:\n", page);
		
		printf("\tLoad.\n");
		//Load the contents into the page buffer
		for(unsigned char word = 0; word < 32; word++)
		{
			//Load low byte of this word
			spi_exchange(0x40);
			spi_exchange(0x00);
			spi_exchange(word);
			spi_exchange(uart_base64_decoded[(page*64)+(word*2)]);
			
			//Load high byte of this word
			spi_exchange(0x48);
			spi_exchange(0x00);
			spi_exchange(word);
			spi_exchange(uart_base64_decoded[(page*64)+(word*2)+1]);
		}
		
		printf("\tWrite.\n");
		
		//Issue the write command
		spi_exchange(0x4C);
		spi_exchange(page >> 3); //(High address bits)
		spi_exchange(page << 5); //(Low address bits)
		spi_exchange(0x00);
		
		printf("\tWait.\n");
		while(1)
		{
			spi_exchange(0xF0);
			spi_exchange(0x00);
			spi_exchange(0x00);
			unsigned char rdybsy = spi_exchange(0x00);
			
			if( (rdybsy & 0x01) == 0 ) //LSB is set to 0 when done
				break;
		}
	}
		
	printf("Program done.\n");
	
	printf("Validating.\n");
	for(unsigned short word = 0; word < 512; word++)
	{
		//Read high byte
		spi_exchange(0x28);
		spi_exchange(word >> 8); //Addr high
		spi_exchange(word & 0xFF); //Addr low
		unsigned char high = spi_exchange(0x00);
		
		//Read low byte
		spi_exchange(0x20);
		spi_exchange(word >> 8); //Addr high
		spi_exchange(word & 0xFF); //Addr low
		unsigned char low = spi_exchange(0x00);
		
		unsigned char expected_low = uart_base64_decoded[ (word*2) ];
		unsigned char expected_high = uart_base64_decoded[ (word*2) + 1];
		
		printf("%.2X %.2X    %.2X %.2X\n", low, high, expected_low, expected_high);
	}

	printf("Checking fuse bits.\n");
	spi_exchange(0x50);
	spi_exchange(0x00);
	spi_exchange(0x00);
	unsigned char fuses = spi_exchange(0x00);
	
	printf("Fuses = 0x%.2X\n", fuses);
	
	printf("Bringing IOCON out of reset.\n");
	
	*pFIO_FLAG_D = 0xFFFF; //All flags high
	ssync();
	for(int nop = 0; nop < 3500000; nop++){	asm("nop;"); }
}

#else //Mockup code follows

#include <stdio.h>
void padloader()
{
	printf("Mockup tried to run padloader but that's nonsensical.\n");
}

#endif // JSYS_PC_MOCKUP switch
