//pads.c
//Handles communication with the IO controller
//Bryan Topp <betopp@betopp.com> 2015

#include <stdio.h>
#include "pads.h"
#include "spr.h"
#include "fram.h"
unsigned char pads_firstplayer = 0;
unsigned char pads_secondplayer = 0;
unsigned char pads_sysbuttons = 0;

#ifndef JSYS_PC_MOCKUP

//Real code here
#include <sys/platform.h>
#include "vdac.h"
#include "spiutils.h"

unsigned char pads_error = 0;

void pads_init()
{
	printf("IO controller check.\n");
	sprintf(spr_textplane[5]+2, "IO controller check.");
	
	printf("Flags as outputs...\n");
	*pFIO_DIR = ~0; //All flags output	
	
	printf("Clearing PF0 - IOCON reset asserted/low\n");
	*pFIO_FLAG_D = 0xFFFE; //PF0;
	
	vdac_fieldwait();
	vdac_fieldwait();
	vdac_fieldwait();
	vdac_fieldwait();
	vdac_fieldwait();
	vdac_fieldwait();
	
	printf("Setting PF0 - IOCON reset deasserted/high\n");
	*pFIO_FLAG_D = 0xFFFF; //PF0;
	
	vdac_fieldwait();
	vdac_fieldwait();
	vdac_fieldwait();
	vdac_fieldwait();
	vdac_fieldwait();
	vdac_fieldwait();
	
	printf("Checking if IOCON is live.\n");	
	
	spi_set_iocon();
	spi_exchange(0xAA);
	for(int nop = 0; nop < 10000; nop++){	asm("nop;"); }
	const unsigned char test_a = spi_exchange(0x00);
	for(int nop = 0; nop < 10000; nop++){	asm("nop;"); }
	spi_exchange(0x55);
	for(int nop = 0; nop < 10000; nop++){	asm("nop;"); }
	const unsigned char test_b = spi_exchange(0x00);
	spi_set_none();
	
	if(test_a != 0x55 || test_b != 0xAA)
	{
		sprintf(spr_textplane[6] + 2, "Error. IOCON unresponsive.");
		fprintf(stderr, "Failed magic, 0xAA -> 0x%.2X, 0x55 -> 0x%.2X\n", test_a, test_b);
		pads_error = 1;
		
		return;
	}
	
	printf("Polling version...\n");
	
	spi_set_iocon();
	spi_exchange(0xEE);
	for(int nop = 0; nop < 10000; nop++){	asm("nop;"); }
	const unsigned char version_byte = spi_exchange(0x00);
	spi_set_none();
	
	printf("IOCON version returns 0x%.2X\n", version_byte);
	sprintf(spr_textplane[6] + 2, "Firmware ver. 0x%.2X", version_byte);

	if(version_byte != 0xC0)
	{
		sprintf(spr_textplane[6] + 22, "BAD VER.");
		fprintf(stderr, "Wrong version, force reload.\n");
		pads_error = 1;
		return;
	}
	
	pads_error = 0;
}


//On the real machine, our IO controller keeps track of coin entries.
//(It runs deterministically and never hitches.)
//So, we ask it for any "pending" coins it's detected.
//It atomically responds and clears them, and we catch and accumulate them here.
unsigned char latched_p1_credits = 0;
unsigned char latched_p2_credits = 0;

void pads_poll()
{
	if(pads_error)
	{
		pads_firstplayer = 0;
		pads_secondplayer = 0;
		pads_sysbuttons = 0;
		return;
	}

	//A bit of shenanigans here, because the IO controller is slow to respond to things
	//So, select the chip, and then slowly poll these, pipelined one-deep
	spi_set_iocon();	
	spi_exchange(0x00);
	for(int nop = 0; nop < 10000; nop++){	asm("nop;"); }	
	
	pads_sysbuttons = spi_exchange(0x11); //Sysbuttons is response from 00 command
	for(int nop = 0; nop < 10000; nop++){	asm("nop;"); }	
	
	pads_firstplayer = spi_exchange(0x22); //First controls is response from 11 command
	for(int nop = 0; nop < 10000; nop++){	asm("nop;"); }	
	
	pads_secondplayer = spi_exchange(0x33); //Second controls is response from 22 command
	for(int nop = 0; nop < 10000; nop++){	asm("nop;"); }	
	
	unsigned char newcoins1 = spi_exchange(0x44); //First pending coinage is response from 33 command
	for(int nop = 0; nop < 10000; nop++){	asm("nop;"); }	
	
	unsigned char newcoins2 = spi_exchange(0x00); //Second pending coinage is response from 44 command
	spi_set_none();
	
	//Update if we got new coins
	latched_p1_credits += newcoins1;
	fram_shadow.coins_deposited_1 += newcoins1;
	
	latched_p2_credits += newcoins2;
	fram_shadow.coins_deposited_2 += newcoins2;
}

unsigned char pads_get_p1_credits()
{
	return latched_p1_credits;
}

unsigned char pads_get_p2_credits()
{
	return latched_p2_credits;
}

unsigned char pads_deduct_p1_credit()
{
	if(latched_p1_credits > 0)
	{
		latched_p1_credits--;
		return 1;
	}
	
	return 0;
}

unsigned char pads_deduct_p2_credit()
{
	if(latched_p2_credits > 0)
	{
		latched_p2_credits--;
		
		return 1;
	}
	
	return 0;
}

unsigned char pads_isvalid()
{
	return pads_error?0:1;
}


#else //JSYS_PC_MOCKUP switch

//Mockup code follows
#include <SDL2/SDL.h>



//On the PC, we just do credit incrementing/decrementing right here.
unsigned char credits_p1 = 0;
unsigned char credits_p2 = 0;

unsigned char pads_get_p1_credits()
{
	return credits_p1;
}

unsigned char pads_get_p2_credits()
{
	return credits_p2;
}

unsigned char pads_deduct_p1_credit()
{
	if(credits_p1 <= 0)
		return 0;
		
	credits_p1--;
	return 1;
}

unsigned char pads_deduct_p2_credit()
{
	if(credits_p2 <= 0)
		return 0;
		
	credits_p2--;
	return 1;
}

void pads_init()
{
	printf("PC mockup. No IO controller.\n");
	sprintf(spr_textplane[5]+2, "PC mockup. No IO controller.");
	sprintf(spr_textplane[6]+2, "SDL ready.");
}

void pads_poll()
{
	const unsigned char *keystates = SDL_GetKeyboardState(NULL);
	
	pads_firstplayer = 0;
	if(keystates[SDL_SCANCODE_LEFT])
		pads_firstplayer |= PADS_BIT_LEFT;
	if(keystates[SDL_SCANCODE_RIGHT])
		pads_firstplayer |= PADS_BIT_RIGHT;
	if(keystates[SDL_SCANCODE_UP])
		pads_firstplayer |= PADS_BIT_UP;
	if(keystates[SDL_SCANCODE_DOWN])
		pads_firstplayer |= PADS_BIT_DOWN;

	if(keystates[SDL_SCANCODE_LCTRL])
		pads_firstplayer |= PADS_BIT_B1;
	if(keystates[SDL_SCANCODE_LALT])
		pads_firstplayer |= PADS_BIT_B2;
	if(keystates[SDL_SCANCODE_SPACE])
		pads_firstplayer |= PADS_BIT_B3;
	if(keystates[SDL_SCANCODE_1])
		pads_firstplayer |= PADS_BIT_START;
	
	pads_secondplayer = 0;
	if(keystates[SDL_SCANCODE_2])
		pads_secondplayer |= PADS_BIT_START;	
	
	pads_sysbuttons = 0;
	if(keystates[SDL_SCANCODE_5])
		pads_sysbuttons |= PADS_SYS_BIT_COIN1;
	if(keystates[SDL_SCANCODE_6])
		pads_sysbuttons |= PADS_SYS_BIT_COIN2;
	if(keystates[SDL_SCANCODE_9])
		pads_sysbuttons |= PADS_SYS_BIT_SERVICE;
	if(keystates[SDL_SCANCODE_F2])
		pads_sysbuttons |= PADS_SYS_BIT_TEST;
	
	
	static int credit_pressed_p1 = 0;
	
	if(pads_sysbuttons & PADS_SYS_BIT_COIN1)
	{
		credit_pressed_p1 = 1;
	}
	else
	{
		if(credit_pressed_p1)
		{
			credits_p1++;
			fram_shadow.coins_deposited_1++;
		}
			
		credit_pressed_p1 = 0;
	}
	
	static int credit_pressed_p2 = 0;
	
	if(pads_sysbuttons & PADS_SYS_BIT_COIN2)
	{
		credit_pressed_p2 = 1;
	}
	else
	{
		if(credit_pressed_p2)
		{
			credits_p2++;
			fram_shadow.coins_deposited_2++;
		}
			
		credit_pressed_p2 = 0;
	}
		
		
}

unsigned char pads_isvalid()
{
	return 1;
}

#endif //JSYS_PC_MOCKUP switch