//shooty.c
//Shooty game
//Bryan Topp <betopp@betopp.com> 2015

#include <stdio.h>
#include "spr.h"
#include "dramfs.h"
#include "vdac.h"
#include "pads.h"
#include "fram.h"
#include "adac.h"
#include <stdlib.h>
#include <string.h>

#include "shooty_common.h"
#include "shooty_fx.h"
#include "shooty_fleet.h"
#include "shooty_player.h"

//Number output by the video DAC to tell us how long the sprite engine took
extern unsigned int vdac_perfstat; 

//PRNG
static unsigned int prng_state = 1337;
unsigned short shooty_prng16()
{
	prng_state += pads_firstplayer + pads_secondplayer;
	prng_state = (1103515245 * prng_state + 12345);
	return prng_state ^ (prng_state >> 16);
}

static void updatestatus(unsigned char num, unsigned char credits, unsigned char blinktimer, char *outloc)
{	
	switch(players[num].state)
	{
		case PLAYER_STATE_INIT:
			if(blinktimer < 16)
			{
				sprintf(outloc, "           ");
			}
			else
			{
				if(credits > 0)
					sprintf(outloc, "Press Start");
				else
					sprintf(outloc, "Insert Coin");
			}
		break;
		case PLAYER_STATE_SPAWNING:
		case PLAYER_STATE_INVULNERABLE:
		case PLAYER_STATE_NORMAL:
		case PLAYER_STATE_EXPLODING:
		case PLAYER_STATE_DEAD:
			if(players[num].extralives > 0)
			{
				sprintf(outloc, "+%d remain  ", players[num].extralives);
			}
			else
			{
				sprintf(outloc, "LAST LIFE  ");
			}
		break;
		case PLAYER_STATE_TYPING:
			sprintf(outloc, "Name? %3.3s  ", players[num].namebuffer);
			if(blinktimer % 8 < 4)
				outloc[6+players[num].namepos] = ' ';
			else if(players[num].namepos >= 3)
				outloc[9] = '_';
		break;
		default:
			sprintf(outloc, "YA BROKE IT");
		break;
	}
}

//Updates the text-layer headers to reflect score, credits, and prompts
static char last_credit_p1 = 0;
static char last_credit_p2 = 0;

static void updateheaders()
{
	unsigned char credit_p1 = pads_get_p1_credits();
	unsigned char credit_p2 = pads_get_p2_credits();
	
	if(credit_p1 > last_credit_p1 || credit_p2 > last_credit_p2)
	{
		unsigned int soundfile = dramfs_findfile("snd/coindrop.sfx");
		adac_play(dramfs_getptr(soundfile), dramfs_getlen(soundfile)/2, 192, 192);
	}
	last_credit_p1 = credit_p1;
	last_credit_p2 = credit_p2;
	
	static unsigned char blinktimer;
	blinktimer--;
	blinktimer = blinktimer&63;
	
	sprintf(&(spr_textplane[1][0]),  "  1Up %7.2u  Hi %7.2u  2Up %7.2u  ", 
		players[0].score, fram_shadow.highscores[0].score, players[1].score);

	char *p1_status_loc = &(spr_textplane[28][2]);
	updatestatus(0, credit_p1, blinktimer, p1_status_loc);
	
	char *p2_status_loc = &(spr_textplane[28][27]);
	updatestatus(1, credit_p2, blinktimer, p2_status_loc);
	
	sprintf(&(spr_textplane[29][0]), "  Credits %-3d              Credits %-3d   ", credit_p1, credit_p2);
}

//This is done commonly - wait for blanking, poll the pads, update the header texts
static void wait_and_poll()
{
	updateheaders();
	vdac_fieldwait();
	pads_poll();
	shooty_prng16();
}

//Clears the screen entirely
static void clean_clear()
{
	vdac_fieldwait();
	spr_sprite_clear();	
	spr_textplane_clear();
}

//Returns nonzero if any players are considered active
static char any_players_active()
{	
	if(players[0].state != PLAYER_STATE_INIT || players[1].state != PLAYER_STATE_INIT)
		return 1;
	
	return 0;
}

//Returns nonzero if either player has deposited enough coins to start the game
static unsigned char start_ready()
{	
	if(pads_get_p1_credits() > 0 || pads_get_p2_credits() > 0)
		return 1;
	
	return 0;
}

//Helpers to wait a number of frames in attract mode
#define ATTRACT_DELAY_FRAME() \
	do \
	{ \
		wait_and_poll(); \
		if(start_ready()) \
			return; \
	} while(0)

#define ATTRACT_DELAY(x) \
	do \
	{ \
		for(unsigned short f = 0; f < x; f++) \
			ATTRACT_DELAY_FRAME(); \
	} while(0)

//...and in game mode
static void game_delay_frame() //This can be a function because it doesn't have to potentially-return
{ 
	wait_and_poll();
	if( (pads_firstplayer & PADS_BIT_START) && (players[0].state == PLAYER_STATE_INIT) )
	{
		if(pads_deduct_p1_credit())
		{
			players[0].score = 0;
			shooty_players_awardlives(0, 3);
		}
	}
	if( (pads_secondplayer & PADS_BIT_START) && (players[1].state == PLAYER_STATE_INIT) )
	{
		if(pads_deduct_p2_credit())
		{
			players[1].score = 0;
			shooty_players_awardlives(1, 3);
		}
	}
}

#define GAME_DELAY(x) \
	do \
	{ \
		for(unsigned short f = 0; f < x; f++) \
			game_delay_frame(); \
	} while(0)	

static void titlescreen()
{
	if(start_ready()) //No attract screens when we're ready to start - fall quickly to start screen
		return;
	
	clean_clear();
	spr_background = dramfs_getptr(dramfs_findfile("bg/attract_title_bg.solid16"));
	spr_sprite_enable = 1;
		
	//Wait at the start
	ATTRACT_DELAY(120);
	
	//"Typewriter" sequence
	{
		const char * const typewriter_text = "Topp Technical Products presents";
		for(short i = 0; i < 32; i++)
		{	
			spr_textplane[5][i+4] = typewriter_text[i];
			ATTRACT_DELAY(2);
		}
	}
	
	//Name of game (todo: graphic logo)
	ATTRACT_DELAY(60);
	sprintf(spr_textplane[7] + 12, "STELLAR STRIKE");
	ATTRACT_DELAY(120); //Let that sink in
	
	//Short instructions
	{
		const char * const typewriter_text = "Destroy ENEMIES for points";
		for(short i = 0; i < 26; i++)
		{	
			spr_textplane[13][i+7] = typewriter_text[i];
			ATTRACT_DELAY(2);
		}
	}
	ATTRACT_DELAY(30);	
	for(unsigned short s = 0; s < 8; s++)
	{
		spr_spritetable[s].data = dramfs_getptr(dramfs_findfile("fleet/fleet_default.8bpp"));
		spr_spritetable[s].h = 16;
		spr_spritetable[s].x = 68 + s*24;
		spr_spritetable[s].y = 80;
		
		ATTRACT_DELAY(4);
	}
	ATTRACT_DELAY(30);
	{
		const char * const typewriter_text = "Look out for special types";
		for(short i = 0; i < 26; i++)
		{	
			spr_textplane[14][i+7] = typewriter_text[i];
			ATTRACT_DELAY(2);
		}
	}
	ATTRACT_DELAY(60);
	
	sprintf(spr_textplane[17]+8, "  GOLD     Bonus points.");
	spr_spritetable[8].data = dramfs_getptr(dramfs_findfile("fleet/fleet_default.8bpp"));
	spr_spritetable[8].h = 16;
	spr_spritetable[8].x = 124;
	spr_spritetable[8].y = 132;
	ATTRACT_DELAY(30);
	
	sprintf(spr_textplane[20]+8, "PURPLE     Drops powerups.");
	spr_spritetable[9].data = dramfs_getptr(dramfs_findfile("fleet/fleet_default.8bpp"));
	spr_spritetable[9].h = 16;
	spr_spritetable[9].x = 124;
	spr_spritetable[9].y = 156;
	ATTRACT_DELAY(30);
	
	sprintf(spr_textplane[23]+8, "   RED     Extra tough.");	
	spr_spritetable[10].data = dramfs_getptr(dramfs_findfile("fleet/fleet_default.8bpp"));
	spr_spritetable[10].h = 16;
	spr_spritetable[10].x = 124;
	spr_spritetable[10].y = 180;
	
	//Wait at the end
	ATTRACT_DELAY(240);
	
}


static void showscores()
{
	if(start_ready())
		return;
	
	clean_clear();
	spr_background = dramfs_getptr(dramfs_findfile("bg/attract_highscore_bg.solid16"));
	
	sprintf(spr_textplane[3] + 12, "- Hall of Fame -");
	
	ATTRACT_DELAY(60);
	
	for(int sc = 0; sc < FRAM_NUM_SCORES; sc++)
	{
		sprintf(spr_textplane[6+(sc*2)] + 10, "%7.2u  %3.3s  #%-2u",
			fram_shadow.highscores[sc].score,
			fram_shadow.highscores[sc].name,
			sc+1);
		
		ATTRACT_DELAY(30);
	}
	
	fram_write();
	ATTRACT_DELAY(600);
}

static void attract()
{
	while(!start_ready())
	{
		//All of these should start returning immediately when we're ready to start
		titlescreen();
		showscores();
	}
}

static void startscreen()
{	
	clean_clear();
	
	spr_background = dramfs_getptr(dramfs_findfile("bg/startscreen_bg.solid16"));
	
	sprintf(spr_textplane[7] + 12, "STELLAR STRIKE");
	
	sprintf(spr_textplane[9] + 13, "PRESS START!");
	
	sprintf(spr_textplane[24] + 3, "(c) Topp Technical Products, 2016");
	
	while(!any_players_active())
	{
		//Somebody needs to sign in.
		game_delay_frame();
	}
	
	unsigned int soundfile = dramfs_findfile("snd/startscreen.sfx");
	adac_play(dramfs_getptr(soundfile), dramfs_getlen(soundfile)/2, 128, 128);
	
	fram_write(); //Nice place to save away new coinage data
}

static void gameover()
{

	clean_clear();
	
	fram_write();		
	
	//Build sprite chain to make background
	spr_background = dramfs_getptr(dramfs_findfile("bg/gameover.solid16"));
	spr_sprite_enable = 1;
	
	for(int d = 0; d < 240; d++)
	{
		wait_and_poll();
		
		//Allow them to skip past the game-over screen by hitting start
		if( (pads_firstplayer & PADS_BIT_START) || (pads_secondplayer & PADS_BIT_START) )
			break;
	}
		
	//Wait for them to release start if they hit it
	//(Todo: make all the other "press start" areas do edge-detection)
	while( (pads_firstplayer & PADS_BIT_START) || (pads_secondplayer & PADS_BIT_START) )
	{
		wait_and_poll();
	}
	
	//Make sure they don't carry any state out of game-over
	memset(&(players[0]), 0, sizeof(players[0]));
	memset(&(players[1]), 0, sizeof(players[1]));
}



static void levelinit(unsigned char level)
{
	clean_clear();	
	
	//Set up background and palette
	spr_background = dramfs_getptr(dramfs_findfile("bg/space0.solid16"));
	spr_loadpalette_bgr(dramfs_getptr(dramfs_findfile("master.pal")));
	spr_sprite_enable = 1;

	//Set up players
	shooty_players_initstage(level);
	
	//Set up fleet
	shooty_fleet_initstage(level);
	
	//Set up FX
	shooty_fx_initstage(level);
}

//Reads the data about players, bullets, and fleet, and updates the sprite table.
static void update_sprites_from_gamedata()
{		
	shooty_players_fillsprites();	
	shooty_fleet_fillsprites();
	shooty_fx_fillsprites();
}

static unsigned char runlevel(unsigned char level)
{
	//Set up game data
	levelinit(level);
	
	//Draw intro text, blink, and hide
	GAME_DELAY(60);
	sprintf(spr_textplane[12] + 16, "Stage %d", level+1);
	GAME_DELAY(30);
	for(unsigned char blink = 0; blink < 3; blink++)
	{
		GAME_DELAY(30);
		sprintf(spr_textplane[13] + 15, "Get ready!");
		GAME_DELAY(30);
		sprintf(spr_textplane[13] + 15, "          ");
	}
	GAME_DELAY(30);
	sprintf(spr_textplane[12] + 16, "        ");
	GAME_DELAY(30);
	
	while(any_players_active())
	{		

		//Compute game
		shooty_players_update();
		shooty_fx_update();
		shooty_fleet_update();
	
		//Counter-stop the score
		if(players[0].score > 9999999)
			players[0].score = 9999999;
		if(players[1].score > 9999999)
			players[1].score = 9999999;
		
		//Wait for blanking
		game_delay_frame();
		
		//Push new sprites
		update_sprites_from_gamedata();
	}
	
	return 0; //Failure
}

static void victory()
{
	
}

static void rungame()
{
	unsigned char lvl;
	for(lvl = 0; lvl < 16; lvl++)
	{
		if(!runlevel(lvl))
			break; //Break at first failure
	}
	
	if(lvl == 16)
		victory();
}

void shooty()
{
	printf("Shooty game to demonstrate BF533 game system\n");
	printf("Bryan Topp <betopp@betopp.com> 2016\n");
	
	spr_loadpalette_bgr(dramfs_getptr(dramfs_findfile("master.pal")));
	
	while(1)
	{
		attract();
		startscreen();
		rungame();
		gameover();
	}
}