//shooty_fleet.c
//Enemy fleet control for shooty game
//Bryan Topp <betopp@betopp.com> 2016

#include <stdio.h>

#include "shooty_common.h"
#include "shooty_fleet.h"
#include "shooty_fx.h"
#include "shooty_player.h"
#include "spr.h"
#include "dramfs.h"
#include "adac.h"

//-----------------
//Enemy Type Definitions
//-----------------

//Types of ships in the enemy fleet
typedef struct fleettype_s {
	void (*behavior_function)(fleetship_t *); //Function pointer for running behavior
	unsigned short pointvalue; //How many points are given, initially, when destroyed
	unsigned short initialhealth; //Initial HP when the ship is spawned
} fleettype_t;

enum FleetTypeIndex
{
	FLEET_TYPE_NONE = 0,
	FLEET_TYPE_CHARGER,
	FLEET_TYPE_MEDUSA,
	FLEET_TYPE_ARC,
	FLEET_TYPE_UFO,
	FLEET_TYPE_TANKER,
	FLEET_TYPE_SWEEPER,
	FLEET_TYPE_ROLLER,
	FLEET_TYPE_MAX
};

fleettype_t fleettypes[FLEET_TYPE_MAX] = {
	
	[FLEET_TYPE_NONE] = {
		.behavior_function = shooty_fleet_behavior_invalid,
		.pointvalue = 0,
		.initialhealth = 0,
	},
	
	[FLEET_TYPE_CHARGER] = { 
		.behavior_function = shooty_fleet_behavior_charger,
		.pointvalue = 100,
		.initialhealth = 200,
	},
	
	[FLEET_TYPE_MEDUSA] = {
		.behavior_function = shooty_fleet_behavior_medusa,
		.pointvalue = 200,
		.initialhealth = 400,
	},	
	
	[FLEET_TYPE_ARC] = {
		.behavior_function = shooty_fleet_behavior_arc,
		.pointvalue = 100,
		.initialhealth = 300,
	},

	[FLEET_TYPE_UFO] = {
		.behavior_function = shooty_fleet_behavior_ufo,
		.pointvalue = 500,
		.initialhealth = 300,
	},
	
	[FLEET_TYPE_TANKER] = {
		.behavior_function = shooty_fleet_behavior_tanker,
		.pointvalue = 500,
		.initialhealth = 500,
	},
	
	[FLEET_TYPE_SWEEPER] = {
		.behavior_function = shooty_fleet_behavior_sweeper,
		.pointvalue = 300,
		.initialhealth = 100,
	},
	
	[FLEET_TYPE_ROLLER] = {
		.behavior_function = shooty_fleet_behavior_roller,
		.pointvalue = 200,
		.initialhealth = 200,
	},
};

//-----------------
//Enemies in Play
//-----------------

//Currently-existing ships
fleetship_t fleetships[FLEET_MAX];

//-----------------
//Spawning and Wave Control
//-----------------

//Spawn data
typedef struct fleetspawn_s {
	short pos_x; //Position, X, to spawn ship
	short pos_y; //Position, Y, to spawn ship
	
	unsigned short type; //Which type to spawn
	
	unsigned char postwait; //How long to wait after, until spawning the next
	unsigned char count; //How many times to spawn before moving on	
} fleetspawn_t;

//Sprite locations
const unsigned char *shooty_fleet_sprite_datas[FLEET_SPRITE_DATA_MAX];

//Formations, each built of a series of spawns
static const fleetspawn_t formation_medusa_line_left[] = {
	{ .pos_x = 80, .pos_y = -16, .type = FLEET_TYPE_MEDUSA, .postwait = 48, .count = 4 },
	{ .count = 0 },
};

static const fleetspawn_t formation_medusa_line_right[] = {
	{ .pos_x = 224, .pos_y = -16, .type = FLEET_TYPE_MEDUSA, .postwait = 48, .count = 4 },
	{ .count = 0 },
};

static const fleetspawn_t formation_charger_sweep_lr[] = {
	{ .pos_x = 148, .pos_y = -16, .type = FLEET_TYPE_CHARGER, .postwait = 20, .count = 1 },
	{ .pos_x = 172, .pos_y = -16, .type = FLEET_TYPE_CHARGER, .postwait = 20, .count = 1 },
	{ .pos_x = 196, .pos_y = -16, .type = FLEET_TYPE_CHARGER, .postwait = 20, .count = 1 },
	{ .pos_x = 220, .pos_y = -16, .type = FLEET_TYPE_CHARGER, .postwait = 20, .count = 1 },
	{ .pos_x = 244, .pos_y = -16, .type = FLEET_TYPE_CHARGER, .postwait = 0, .count = 1 },
	{ .count = 0 },
};

static const fleetspawn_t formation_charger_sweep_rl[] = {
	{ .pos_x = 156, .pos_y = -16, .type = FLEET_TYPE_CHARGER, .postwait = 20, .count = 1 },
	{ .pos_x = 132, .pos_y = -16, .type = FLEET_TYPE_CHARGER, .postwait = 20, .count = 1 },
	{ .pos_x = 108, .pos_y = -16, .type = FLEET_TYPE_CHARGER, .postwait = 20, .count = 1 },
	{ .pos_x = 84,  .pos_y = -16, .type = FLEET_TYPE_CHARGER, .postwait = 20, .count = 1 },
	{ .pos_x = 60,  .pos_y = -16, .type = FLEET_TYPE_CHARGER, .postwait = 0, .count = 1 },
	{ .count = 0 },
};

static const fleetspawn_t formation_arcs[] = {
	{ .pos_x = 244, .pos_y = -16, .type = FLEET_TYPE_ARC, .postwait = 0, .count = 1 },	
	{ .pos_x = 60,  .pos_y = -16, .type = FLEET_TYPE_ARC, .postwait = 0, .count = 1 },
	{ .count = 0 }
};

static const fleetspawn_t formation_arcwave[] = {
	{ .pos_x = 20, .pos_y = -16, .type = FLEET_TYPE_ARC, .postwait = 0, .count = 1 },	
	{ .pos_x = 308,  .pos_y = -16, .type = FLEET_TYPE_ARC, .postwait = 0, .count = 1 },
	{ .pos_x = 60,  .pos_y = -16, .type = FLEET_TYPE_ARC, .postwait = 0, .count = 1 },
	{ .pos_x = 268,  .pos_y = -16, .type = FLEET_TYPE_ARC, .postwait = 0, .count = 1 },
	{ .pos_x = 100,  .pos_y = -16, .type = FLEET_TYPE_ARC, .postwait = 0, .count = 1 },
	{ .pos_x = 228,  .pos_y = -16, .type = FLEET_TYPE_ARC, .postwait = 0, .count = 1 },
	{ .count = 0 }
};

static const fleetspawn_t formation_wait[] = {
	{ .pos_x = -32, .pos_y = 255, .type = FLEET_TYPE_NONE, .postwait = 127, .count = 1},
	{ .count = 0 }
};

static const fleetspawn_t formation_ufos[] = {
	{ .pos_x = 128, .pos_y = -16, .type = FLEET_TYPE_UFO, .postwait = 0, .count = 1 },	
	{ .pos_x = 176,  .pos_y = -16, .type = FLEET_TYPE_UFO, .postwait = 0, .count = 1 },
	{ .count = 0 }
};

static const fleetspawn_t formation_sweepers_l[] = {
	{ .pos_x = 8, .pos_y = -16, .type = FLEET_TYPE_SWEEPER, .postwait = 0, .count = 1 },	
	{ .pos_x = 8, .pos_y = -32, .type = FLEET_TYPE_SWEEPER, .postwait = 0, .count = 1 },
	{ .pos_x = 8, .pos_y = -48, .type = FLEET_TYPE_SWEEPER, .postwait = 0, .count = 1 },
	{ .pos_x = 8, .pos_y = -64, .type = FLEET_TYPE_SWEEPER, .postwait = 0, .count = 1 },
	{ .count = 0 }
};

static const fleetspawn_t formation_sweepers_r[] = {
	{ .pos_x = 296, .pos_y = -16, .type = FLEET_TYPE_SWEEPER, .postwait = 0, .count = 1 },	
	{ .pos_x = 296, .pos_y = -32, .type = FLEET_TYPE_SWEEPER, .postwait = 0, .count = 1 },
	{ .pos_x = 296, .pos_y = -48, .type = FLEET_TYPE_SWEEPER, .postwait = 0, .count = 1 },
	{ .pos_x = 296, .pos_y = -64, .type = FLEET_TYPE_SWEEPER, .postwait = 0, .count = 1 },
	{ .count = 0 }
};

static const fleetspawn_t formation_tanker[] = {
	{ .pos_x = 152, .pos_y = -16, .type = FLEET_TYPE_TANKER, .postwait = 0, .count = 1 },	
	{ .count = 0 }
};

static const fleetspawn_t formation_rollers_bothsides[] = {
	{ .pos_x = 8, .pos_y = -16, .type = FLEET_TYPE_ROLLER, .postwait = 0, .count = 1 },	
	{ .pos_x = 296, .pos_y = -16, .type = FLEET_TYPE_ROLLER, .postwait = 0, .count = 1 },	
	{ .count = 0 }
};

//Formations table, currently just run in-order
static const fleetspawn_t *formations_table[] = {
	formation_tanker,
	formation_rollers_bothsides,
	formation_wait,
	formation_arcs,
	formation_ufos,
	formation_wait,
	formation_sweepers_l,
	formation_wait,
	formation_sweepers_r,
	formation_wait,
	formation_medusa_line_left,
	formation_medusa_line_right,
	formation_wait,
	formation_arcwave,
	formation_tanker,
	formation_wait,
	formation_charger_sweep_lr,
	formation_wait,
	formation_charger_sweep_rl,
	formation_wait,
	NULL
};

//Tracking spawns
static short fleet_spawntimer;
static short fleet_spawnrepeat;
static short fleet_formation;
static short fleet_formation_step;


//Possibly spawns the next fleet ship
static void check_fleet_spawn()
{
	//If we still have time remaining, count down and return
	if(fleet_spawntimer > 0)
	{
		fleet_spawntimer--;
		return;
	}

	//Alright, time to spawn. Do we need to move ahead in the table, or merely repeat?
	if(fleet_spawnrepeat == 0)
	{
		//Next entry - increment "which" pointer, and grab the appropriate number of total repetitions
		fleet_formation_step++;
		if(formations_table[fleet_formation][fleet_formation_step].count == 0) //Done with this formation, move on
		{
			fleet_formation_step = 0;
			fleet_formation++;
			if(formations_table[fleet_formation] == 0)
				fleet_formation = 0; //Loop
		}
		
		fleet_spawnrepeat = formations_table[fleet_formation][fleet_formation_step].count;
	}
	
	//Find an empty spot to spawn in
	unsigned char s;
	for(s = 0; s < FLEET_MAX; s++)
	{
		if(fleetships[s].pos_y >= 240)
			break;
	}
	
	if(s >= FLEET_MAX) //Couldn't find a free spot - try next frame until we can
		return;
		
	
	//Knock off a repetition, restart spawn timer
	fleet_spawnrepeat--;
	fleet_spawntimer = formations_table[fleet_formation][fleet_formation_step].postwait;

	//Spawn them
	fleetships[s].pos_x = formations_table[fleet_formation][fleet_formation_step].pos_x;
	fleetships[s].pos_y = formations_table[fleet_formation][fleet_formation_step].pos_y;
	fleetships[s].type = formations_table[fleet_formation][fleet_formation_step].type;
	fleetships[s].behavior_step = 0;
	
	fleetships[s].health = fleettypes[fleetships[s].type].initialhealth;
}

//Prepare for new stage
void shooty_fleet_initstage(unsigned char level)
{
	//Cache locations of fleet sprites from DRAMFS
	for(unsigned char fsd = 0; fsd < FLEET_SPRITE_DATA_MAX; fsd++)
	{
		shooty_fleet_sprite_datas[fsd] = dramfs_getptr(dramfs_findfile(shooty_fleet_sprite_data_names[fsd]));
	}
	
	//Set up sprite table
	for(unsigned char s = SIDX_FLEET; s <= SIDX_FLEET_LAST; s++)
	{
		spr_spritetable[s].data = shooty_fleet_sprite_datas[FLEET_SPRITE_DATA_DEFAULT];
		spr_spritetable[s].h = 16;
		spr_spritetable[s].x = -24;
		spr_spritetable[s].y = 255;
		
		fleetships[s-SIDX_FLEET].pos_x = -24;
		fleetships[s-SIDX_FLEET].pos_y = 255;
		fleetships[s-SIDX_FLEET].desired_sprite = FLEET_SPRITE_DATA_DEFAULT;
	}
	
	//Set up spawner
	fleet_spawntimer = 60;
	fleet_spawnrepeat = 0;
	fleet_formation = 0;
	fleet_formation_step = -1;
	
}

//Moves the fleet ships around; possibly spawns new fleet ships or fleet bullets
void shooty_fleet_update()
{
	//Move existing fleep ships
	for(unsigned char s = 0; s < FLEET_MAX; s++)
	{
		//Dead ships are moved off screen
		if(fleetships[s].pos_y > 240)
			continue;
		
		//Otherwise, continue behavior
		fleettypes[fleetships[s].type].behavior_function(&(fleetships[s]));
		
		//Catch ships which have moved far off-screen
		if(fleetships[s].pos_x < -128)
			fleetships[s].pos_y = 255;
		if(fleetships[s].pos_x > 320 + 128)
			fleetships[s].pos_y = 255;
	}
	
	//See if we need to spawn more - and do so
	check_fleet_spawn();
}

//Fills sprite data
void shooty_fleet_fillsprites()
{
	for(unsigned char f = 0; f < FLEET_MAX; f++)
	{
		spr_spritetable[SIDX_FLEET+f].x = fleetships[f].pos_x;
		spr_spritetable[SIDX_FLEET+f].y = fleetships[f].pos_y;
		
		spr_spritetable[SIDX_FLEET+f].data = shooty_fleet_sprite_datas[fleetships[f].desired_sprite];
	}
}

//Damages and possibly destroys a ship in the fleet
void shooty_fleet_damage(unsigned char which, unsigned short amt, unsigned char causing_player)
{
	if(which >= FLEET_MAX)
		return;
	
	fleetships[which].health -= amt;
	if(fleetships[which].health <= 0)
	{
		//Play explosion sound
		unsigned int soundfile = dramfs_findfile("snd/explosion.sfx");
		short pan = (fleetships[which].pos_x - 160) / 2;
		adac_play(dramfs_getptr(soundfile), dramfs_getlen(soundfile)/2, 128 - pan, 128 + pan);

		//Award points and show points FX
		unsigned short pointvalue = fleettypes[fleetships[which].type].pointvalue;
		shooty_players_awardpts(causing_player, pointvalue);
		shooty_fx_spawn_points_helper(fleetships[which].pos_x, fleetships[which].pos_y, pointvalue);
		
		//Spawn the death FX and move them out of the way
		shooty_fx_spawn(fleetships[which].pos_x, fleetships[which].pos_y, FXTYPE_DEATH);
		fleetships[which].pos_y = 255;
		fleetships[which].pos_x = -24;
		fleetships[which].type = FLEET_TYPE_NONE;
		fleetships[which].desired_sprite = FLEET_SPRITE_DATA_DEFAULT;
		
	}
}
