//shooty_fleet_behaviors.c
//Enemy behavior update functions
//Bryan Topp <betopp@betopp.com> 2016

#include "shooty_fleet.h"
#include "shooty_player.h"
#include "spr.h"
#include "dramfs.h"

//"Default" behavior when something goes wrong - kick off playing field
void shooty_fleet_behavior_invalid(fleetship_t *ship)
{
	ship->pos_y = 255;
}

//Charger - advances in, stops for a moment, aims at the player (roughly), and then rushes at (and past) them
void shooty_fleet_behavior_charger(fleetship_t *ship)
{
	if(ship->behavior_step == 0)
	{
		//Step 0 - Moving in initially
		ship->desired_sprite = FLEET_SPRITE_DATA_CHARGER_S;
		
		if(ship->pos_y < 16)
			ship->pos_y += 3;
		else if(ship->pos_y < 32)
			ship->pos_y += 2;
		else
		{
			ship->pos_y++;
			if(shooty_prng16() % 1024 < ship->pos_y || ship->pos_y > 128)
				ship->behavior_step = 1;
		}
	}
	else if(ship->behavior_step == 1)
	{		
		//Step 1 - Picking a player to aim at
		unsigned char preferred_player = shooty_prng16()%2;
		if(players[preferred_player].state < PLAYER_STATE_SPAWNING || players[preferred_player].state > PLAYER_STATE_EXPLODING)
			preferred_player ^= 1;
		
		ship->behavior_step = 2 + preferred_player;
	}
	else if(ship->behavior_step < 64)
	{
		//Steps less than 64, taking aim at a player (even = first player, odd = second player)
		short xdiff = players[ship->behavior_step%2].pos_x - ship->pos_x;
		short ydiff = players[ship->behavior_step%2].pos_y - ship->pos_y;
		
		unsigned char aimdir;
		if(ydiff <= 0) //Player above
		{
			if(xdiff < 0)
				aimdir = 0;
			else
				aimdir = 4;
		}
		else
		{
			if(2 * xdiff > ydiff)
				aimdir = 4;
			else if(4 * xdiff > ydiff)
				aimdir = 3;
			else if(2 * xdiff < -ydiff)
				aimdir = 0;
			else if(4 * xdiff < -ydiff)
				aimdir = 1;
			else
				aimdir = 2;
		}
		
		switch(aimdir)
		{
			case 0:
				ship->desired_sprite = FLEET_SPRITE_DATA_CHARGER_SW;
				break;
			case 1:
				ship->desired_sprite = FLEET_SPRITE_DATA_CHARGER_SSW;
				break;
			case 3:
				ship->desired_sprite = FLEET_SPRITE_DATA_CHARGER_SSE;
				break;
			case 4:
				ship->desired_sprite = FLEET_SPRITE_DATA_CHARGER_SE;
				break;
			default:
				ship->desired_sprite = FLEET_SPRITE_DATA_CHARGER_S;
				break;
		}
		
		//After quarter a second, start vibrating
		if(ship->behavior_step > 30)
		{
			if(ship->behavior_step & 0x4) //Every two steps, as we advance by two each time
				ship->pos_y+=2;
			else
				ship->pos_y-=2;
		}
		
		ship->behavior_step += 2; //Maintain even/odd to keep aiming at same player
		
		//Move onto a charge after a half second (remember, advancing by two)
		if(ship->behavior_step >= 64)
			ship->behavior_step = 64 + aimdir;
	}
	else if(ship->behavior_step == 64)
	{
		//Charging SW
		ship->pos_y += 2;
		ship->pos_x -= 2;
	}
	else if(ship->behavior_step == 65)
	{
		//Charging SSW
		ship->pos_y += 3;
		ship->pos_x -= 1;
	}
	else if(ship->behavior_step == 66)
	{
		//Charging S
		ship->pos_y += 3;
	}
	else if(ship->behavior_step == 67)
	{
		//Charging SSE
		ship->pos_y += 3;
		ship->pos_x += 1;
	}
	else if(ship->behavior_step == 68)
	{
		//Charging SE
		ship->pos_y += 2;
		ship->pos_x += 2;
	}
	else
	{
		ship->pos_y = 255;
	}
}

static const char sintab[64] = 
{
	0,	2,	5,	7,	9,	11,	13,	15,
	17, 	19, 	20, 	21, 	22, 	23, 	24, 	24, 
	24, 	24, 	24,	23,	22,	21, 	20, 	19, 	
	17, 	15, 	13, 	11,	9,	7,	5, 	2, 
	0, 	-2, 	-5,	-7, 	-9, 	-11,	-13, 	-15, 
	-17,	-19,	-20, 	-21,	-22, 	-23, 	-24,	-24,
	-24, 	-24,	-24,	-23,	-22,	-21,	-20,	-19,
	-17,	-15,	-13,	-11,	-9, 	-7, 	-5,	-2
};

//Medusa (named after the CV enemy) - Floats left/right in a sinusoid pattern	
void shooty_fleet_behavior_medusa(fleetship_t *ship)
{	
	ship->desired_sprite = FLEET_SPRITE_DATA_MEDUSA;
	ship->pos_y ++;
	
	ship->pos_x -= sintab[(ship->behavior_step++)%64];
	ship->pos_x += sintab[(ship->behavior_step)%64];
}

//Arc - Starts attacking the player, but then runs away
void shooty_fleet_behavior_arc(fleetship_t *ship)
{
	ship->desired_sprite = FLEET_SPRITE_DATA_ARC;

	if(ship->behavior_step == 0)
	{
		//Check for playrs close enough to spook
		for(unsigned char p = 0; p < 2; p++)
		{
			if(players[p].state < PLAYER_STATE_SPAWNING || players[p].state > PLAYER_STATE_EXPLODING)
				continue; //Player is dead/gone/etc
			
			if(players[p].pos_y < ship->pos_y)
				continue; //Player is above us (at all)
			
			if(players[p].pos_y > ship->pos_y + 32)
				continue; //Player too far below us
			
			if(players[p].pos_x < ship->pos_x - 32)
				continue; //Player too far to our left
			
			if(players[p].pos_x > ship->pos_x + 32)
				continue; //Player too far to our right
			
			//Spook!
			ship->behavior_step = 16;
		}		
		
		if(ship->pos_y < 40)
			ship->pos_y += 3;
		else
			ship->pos_y += 2;
	}
	else if(ship->behavior_step <= 16)
	{
		//Spooked		
		ship->pos_y -= ship->behavior_step / 2; //Back up in decaying arc

		if(ship->behavior_step % 2)
			ship->pos_x += (ship->behavior_step & 2)?2:-2;
		
		ship->behavior_step--;
		
		if(ship->behavior_step == 0)
			ship->behavior_step = 100;
	}
	else
	{
		//CHARGE!
		ship->pos_y += 4;
	}
		
}

//UFO - Moves horizontally across the screen, wobbling
void shooty_fleet_behavior_ufo(fleetship_t *ship)
{
	ship->desired_sprite = FLEET_SPRITE_DATA_UFO;
	
	if(ship->pos_y < 64 && ship->behavior_step == 0) //Too close to top of screen, drop
	{
		ship->pos_y+=2;
	}
	else //Ready to hover
	{
		ship->pos_y -= sintab[(ship->behavior_step++)%64];
		ship->pos_y += sintab[(ship->behavior_step)%64];
		
		if(ship->pos_x < 160)
			ship->pos_x--;
		else
			ship->pos_x++;
	}
}

//Tanker - plods slowly, accelerating then decaying vertically
void shooty_fleet_behavior_tanker(fleetship_t *ship)
{
	ship->desired_sprite = FLEET_SPRITE_DATA_TANKER;
	
	unsigned char vvel = (ship->behavior_step & 0x001F) >> 0; //5 bits of velocity
	unsigned char subpx = (ship->behavior_step & 0x00E0) >> 5; //3 bits of subpixel
	unsigned char step = (ship->behavior_step & 0x3F00) >> 8; //6 bits of step
	
	if(vvel > 2 && (step % 2))
		vvel--;
	
	if(step == 0) //Burst periodically
		vvel = 16;
	
	subpx += vvel; //Accumulate velocity into subpixels
	ship->pos_y += subpx >> 3; //Overflow from subpixels goes to actual position	
	step++; //Step increments
	
	ship->behavior_step = (vvel & 0x1f) | ((subpx & 0x07) << 5) | ((step & 0xFF) << 8);
	
}

//Sweeper - moves in vertically, then rushes across horizontally
void shooty_fleet_behavior_sweeper(fleetship_t *ship)
{
	if(ship->behavior_step < 64)
	{
		//Moving in vertically
		ship->pos_y += (64 - ship->behavior_step++) / 8;
		ship->desired_sprite = FLEET_SPRITE_DATA_SWEEPER_S;
	}
	else if(ship->behavior_step == 64)
	{
		//Done moving in - pick aim direction
		if(ship->pos_x < 152)
		{
			ship->behavior_step = 65;
			ship->desired_sprite = FLEET_SPRITE_DATA_SWEEPER_E;
		}
		else
		{
			ship->behavior_step = 66;
			ship->desired_sprite = FLEET_SPRITE_DATA_SWEEPER_W;
		}
	}
	else if(ship->behavior_step < 128)
	{
		//Holding position, waiting to sweep
		ship->behavior_step+=2;
		
		if(ship->behavior_step & 2)
			ship->pos_y += (ship->behavior_step & 4)?1:-1;
	}		
	else if(ship->behavior_step % 2)
	{
		//Charging right
		ship->desired_sprite = FLEET_SPRITE_DATA_SWEEPER_E;
		ship->pos_x += 4;
	}
	else
	{
		//Charging left
		ship->desired_sprite = FLEET_SPRITE_DATA_SWEEPER_W;
		ship->pos_x -= 4;
	}
}

//Roller - bounces along the side of the screen
void shooty_fleet_behavior_roller(fleetship_t *ship)
{
	ship->desired_sprite = FLEET_SPRITE_DATA_ROLLER;
	
	if(ship->behavior_step < 16)
	{
		//"Falling" to side
		if(ship->pos_x < 152)
		{
			ship->pos_x -= ship->behavior_step / 4;
			if(ship->pos_x <= 0)
			{
				ship->pos_x = 0;
				ship->behavior_step = 17;
			}
			else
			{
				if(ship->behavior_step < 15)
					ship->behavior_step++;
			}
		}
		else
		{
			ship->pos_x += ship->behavior_step / 4;
			if(ship->pos_x >= 304)
			{
				ship->pos_x = 304;
				ship->behavior_step = 17;
			}
			else
			{
				if(ship->behavior_step < 15)
					ship->behavior_step++;
			}
		}
	}
	else
	{
		//Bouncing away from side
		if(ship->pos_x < 152)
		{
			ship->pos_x += (32-ship->behavior_step) / 4;
		}
		else
		{
			ship->pos_x -= (32-ship->behavior_step) / 4;
		}
		
		ship->behavior_step++;
		if(ship->behavior_step >= 32)
			ship->behavior_step = 0;
	}
	
	ship->pos_y+=2;
}
