//shooty_player.c
//Player ships, lives, scoring
//Bryan Topp <betopp@betopp.com> 2016

#include "shooty_common.h"
#include "shooty_player.h"
#include "shooty_fx.h"
#include "shooty_fleet.h"
#include "spr.h"
#include "dramfs.h"
#include "pads.h"
#include "fram.h"
#include "adac.h"


playerstate_t players[2];
playerbullet_t playerbullets[2][BULLET_MAX];


//Spawns a new bullet from the specified player's ship
static void fire_bullet(unsigned char num)
{
	short startpos_x = players[num].pos_x;
	short startpos_y = players[num].pos_y;
	
	unsigned char b;
	for(b = 0; b < BULLET_MAX; b++)
	{
		if(playerbullets[num][b].pos_y < -16) //Unused bullets are off-screen
			break;
	}
	
	if(b >= BULLET_MAX)
		return; //Can't fire more bullets right now
	
	playerbullets[num][b].pos_x = startpos_x; // - 1 + ((players[num].stateframe%2)*2);
	playerbullets[num][b].pos_y = startpos_y;
	
	unsigned int soundfile = dramfs_findfile("snd/pshoot.sfx");
	short pan = (playerbullets[num][b].pos_x - 160) / 4;
	adac_play(dramfs_getptr(soundfile), dramfs_getlen(soundfile)/2, 64 - pan, 64 + pan);
}


//Moves the specified player's position and possibly fires bullets based on their controls
static void update_player_control(unsigned char num)
{
	playerstate_t *p = &players[num];
	unsigned char input = num?pads_secondplayer:pads_firstplayer;

	//Track how long they've held the shoot button.
	if( (input & PADS_BIT_B1) && (p->holdingshoot < 255) )
		p->holdingshoot++;
	
	if( !(input & PADS_BIT_B1) )
		p->holdingshoot = 0;
	
	//Move the player
	if(input & PADS_BIT_LEFT)
		p->pos_x-=2;
	if(input & PADS_BIT_RIGHT)
		p->pos_x+=2;
	
	if(input & PADS_BIT_UP)
		p->pos_y-=1;
	if(input & PADS_BIT_DOWN)
		p->pos_y+=1;
	
	//Cap at screen boundaries
	if(p->pos_x < 0)
		p->pos_x = 0;
	if(p->pos_x > 304)
		p->pos_x = 304;
	
	if(p->pos_y < 0)
		p->pos_y = 0;
	if(p->pos_y > 224)
		p->pos_y = 224;
	
	//Queue shots each time they hit fire
	if( p->holdingshoot == 1 )
	{
		p->shotqueue = 2;
	}
	
	//Drain shot queue - fire bullets
	if(p->shotcooldown > 0)
	{
		p->shotcooldown--;
	}
	else if(p->shotqueue > 0)
	{
		p->shotcooldown = 6;
		p->shotqueue--;
		
		fire_bullet(num);
	}
}



//Checks for collisions with the specified player.
//Returns true and deals damage to the enemy if one is found
static char check_player_hit(unsigned char num)
{
	playerstate_t *p = &players[num];
	
	for(unsigned char fidx = 0; fidx < FLEET_MAX; fidx++)
	{
		fleetship_t *f = &fleetships[fidx];
		
		if(p->pos_x < f->pos_x - 10)
			continue;
		if(p->pos_x > f->pos_x + 10)
			continue;
		if(p->pos_y < f->pos_y - 10)
			continue;
		if(p->pos_y > f->pos_y + 10)
			continue;
		
		shooty_fleet_damage(fidx, 1000, num);
		return 1;
	}
	
	return 0;
}


//Moves active bullets for the specified player; damage any enemies if hit
static void update_player_bullets(unsigned char num)
{
	for(unsigned char b = 0; b < BULLET_MAX; b++)
	{
		if(playerbullets[num][b].pos_y > -24)
		{
			playerbullets[num][b].pos_y -= 8;
			
			for(unsigned char f = 0; f < FLEET_MAX; f++)
			{
				if(playerbullets[num][b].pos_y < fleetships[f].pos_y - 8)
					continue;
				if(playerbullets[num][b].pos_y > fleetships[f].pos_y + 8)
					continue;
				
				if(playerbullets[num][b].pos_x > fleetships[f].pos_x + 10)
					continue;
				if(playerbullets[num][b].pos_x < fleetships[f].pos_x - 10)
					continue;
				
				//Hit!
				
				unsigned int soundfile = dramfs_findfile("snd/hit.sfx");
				short pan = (playerbullets[num][b].pos_x - 160) / 2;
				adac_play(dramfs_getptr(soundfile), dramfs_getlen(soundfile)/2, 128 - pan, 128 + pan);
				
				shooty_fx_spawn(playerbullets[num][b].pos_x, playerbullets[num][b].pos_y, FXTYPE_HIT);
				
				playerbullets[num][b].pos_y = -24; //Bullet has expired
				
				shooty_fleet_damage(f, 100, num); //Damage target
			}
		}
	}
}

//Updates the specified player based on their current state
static void update_player(unsigned char num)
{
	playerstate_t *p = &players[num];
	
	//Increment frame count
	p->stateframe++;
	
	//Switch based on their state
	switch(p->state)
	{
		case PLAYER_STATE_INIT:
		{
			//Make sure they're offscreen
			p->pos_x = -24;
			p->pos_y = -24;
			
			//Game starts by adding some extra lives - permit leaving init, in that case
			if(p->extralives > 0)
			{
				p->state = PLAYER_STATE_SPAWNING;
				p->stateframe = 0;
				p->extralives--;
			}
				
			break;
		}
		case PLAYER_STATE_SPAWNING:
		{
			//Play sound when we start to spawn
			if(p->stateframe == 1)
			{
				unsigned int soundfile = dramfs_findfile("snd/spawn.sfx");
				adac_play(dramfs_getptr(soundfile), dramfs_getlen(soundfile)/2, 128, 128);
			}
			
			//Animate in based on how long we've been in this state
			p->pos_x = num?168:136;
			p->pos_y = 240 - (p->stateframe/2);
			
			if(p->stateframe > 96) //Once they've slid on-screen, let them move around while invulnerable
			{
				p->state = PLAYER_STATE_INVULNERABLE;
				p->stateframe = 0;
			}
			
			break;
		}
		case PLAYER_STATE_INVULNERABLE:
		{
			//Let them move around and shoot, but don't process collisions against them
			update_player_control(num);
			
			if(p->stateframe > 120) //2 seconds of being invulnerable
			{
				p->state = PLAYER_STATE_NORMAL;
				p->stateframe = 0;
			}
			
			break;
		}
		case PLAYER_STATE_NORMAL:
		{
			//Main procedure
			update_player_control(num);
			
			if(check_player_hit(num))
			{
				p->state = PLAYER_STATE_EXPLODING;
				p->stateframe = 0;
			}
			
			break;
		}
		case PLAYER_STATE_EXPLODING:
		{
			
			p->shotqueue = 0;
			p->shotcooldown = 0;
			p->holdingshoot = 0;
			
			if(p->stateframe == 1)
			{				
				shooty_fx_spawn(p->pos_x + (shooty_prng16()%8) - 4, p->pos_y + (shooty_prng16()%8) - 4, FXTYPE_DEATH);
			}
			else if(p->stateframe < 60 && (p->stateframe % 4 == 0))
			{
				shooty_fx_spawn(p->pos_x + (shooty_prng16()%16) - 8, p->pos_y + (shooty_prng16()%16) - 8, FXTYPE_HIT);
			}
			else if(p->stateframe > 120)
			{
				p->pos_y = -24;
				p->state = PLAYER_STATE_DEAD;
				p->stateframe = 0;
			}
			
			
			break;
		}
		case PLAYER_STATE_DEAD:
		{
			//What happens now depends on their extra lives...
			if(p->extralives > 0) //More lives - deduct and start spawning
			{
				p->state = PLAYER_STATE_SPAWNING;
				p->stateframe = 0;
				p->extralives--;
			}
			else //No more lives - let them type their high score
			{
				p->state = PLAYER_STATE_TYPING;
				p->stateframe = 0;
				
				//Initialize high-score initial entry
				p->namebuffer[0] = 'A';
				p->namebuffer[1] = 'A';
				p->namebuffer[2] = 'A';
				p->namepos = 0;
				p->namedebounce = 0;
			}
			break;			
		}
		case PLAYER_STATE_TYPING:
		{
			unsigned char input = num?pads_secondplayer:pads_firstplayer;
						
			if(p->namedebounce > 0)
			{
				if(input == 0)
					p->namedebounce = 0;
			}
			else
			{		
				if(input != 0)
					p->namedebounce = 1;
				
				if(p->namepos > 0 && (input & PADS_BIT_LEFT))
				{
					//Not at the beginning, and hitting left. Back up.
					p->namepos--;
				}
				else if(p->namepos < 3 && (input & (PADS_BIT_RIGHT | PADS_BIT_B1)))
				{
					//On one of the characters, and hitting right or A. Advance.
					p->namepos++;
				}
				else if(p->namepos == 3 && (input & PADS_BIT_B1))
				{
					//Past the end of the name (i.e. index 3), hit A. Finish.
					p->namepos++;
				}
				else if(p->namepos >= 0 && p->namepos <= 2)
				{
					//In the name - let them edit it
					if(input & PADS_BIT_UP)
						p->namebuffer[p->namepos]++;
					if(input & PADS_BIT_DOWN)
						p->namebuffer[p->namepos]--;
					
					//Wrap
					if(p->namebuffer[p->namepos] < ' ')
						p->namebuffer[p->namepos] = '_';
					if(p->namebuffer[p->namepos] > '_')
						p->namebuffer[p->namepos] = ' ';
				}
			}
			
			//If they're done typing, or timed out, commit and send them back to init
			if( p->stateframe > 3600 || p->namepos >= 4)
			{
				
				fram_insert_highscore(p->score, p->namebuffer, 1);
				
				p->state = PLAYER_STATE_INIT;
				p->stateframe = 0;
				p->namepos = 0;
			}
			
			break;
		}
		default:
		{
			p->state = PLAYER_STATE_INIT;
			p->stateframe = 0;
			break;
		}
	}
}

void shooty_players_fillsprites()
{
	//Set new sprites for players
	spr_spritetable[SIDX_P0].x = players[0].pos_x;
	spr_spritetable[SIDX_P0].y = players[0].pos_y;
	spr_spritetable[SIDX_P1].x = players[1].pos_x;
	spr_spritetable[SIDX_P1].y = players[1].pos_y;
	
	//Blink players if they're invulnerable
	if(players[0].stateframe % 2 && players[0].state < PLAYER_STATE_NORMAL)
		spr_spritetable[SIDX_P0].y = -24;
	if(players[1].stateframe % 2 && players[1].state < PLAYER_STATE_NORMAL)
		spr_spritetable[SIDX_P1].y = -24;
	
	//Make them invisible if they're exploding/dead
	if(players[0].state > PLAYER_STATE_NORMAL)
		spr_spritetable[SIDX_P0].y = -24;
	if(players[1].state > PLAYER_STATE_NORMAL)
		spr_spritetable[SIDX_P1].y = -24;
	
	//Set sprite for bullets
	for(unsigned char b = 0; b < BULLET_MAX; b++)
	{
		spr_spritetable[SIDX_P0_BULLETS+b].x = playerbullets[0][b].pos_x;
		spr_spritetable[SIDX_P0_BULLETS+b].y = playerbullets[0][b].pos_y;
		
		spr_spritetable[SIDX_P1_BULLETS+b].x = playerbullets[1][b].pos_x;
		spr_spritetable[SIDX_P1_BULLETS+b].y = playerbullets[1][b].pos_y;
	}
}

void shooty_players_initstage(unsigned char level)
{
	//Set up player sprites
	spr_spritetable[SIDX_P0].data = dramfs_getptr(dramfs_findfile("player/player_default.8bpp"));
	spr_spritetable[SIDX_P0].h = 16;
	spr_spritetable[SIDX_P0].x = -24;
	spr_spritetable[SIDX_P0].y = -24;
	
	spr_spritetable[SIDX_P1].data = dramfs_getptr(dramfs_findfile("player/player_default.8bpp"));
	spr_spritetable[SIDX_P1].h = 16;
	spr_spritetable[SIDX_P1].x = -24;
	spr_spritetable[SIDX_P1].y = -24;

	//Set up bullets
	for(unsigned char s = SIDX_P0_BULLETS; s <= SIDX_P0_BULLETS_LAST; s++)
	{
		spr_spritetable[s].data = dramfs_getptr(dramfs_findfile("player/bullet_default.8bpp"));
		spr_spritetable[s].h = 16;
		spr_spritetable[s].x = -24;
		spr_spritetable[s].y = -24;
		
		playerbullets[0][s-SIDX_P0_BULLETS].pos_y = -24;
	}
	
	for(unsigned char s = SIDX_P1_BULLETS; s <= SIDX_P1_BULLETS_LAST; s++)
	{
		spr_spritetable[s].data = dramfs_getptr(dramfs_findfile("player/bullet_default.8bpp"));
		spr_spritetable[s].h = 16;
		spr_spritetable[s].x = -24;
		spr_spritetable[s].y = -24;
		
		playerbullets[1][s-SIDX_P1_BULLETS].pos_y = -24;
	}
}

void shooty_players_update()
{
	for(unsigned char p = 0; p < 2; p++)
	{
		update_player(p);	
		update_player_bullets(p);
	}
}

void shooty_players_awardpts(unsigned char which, unsigned short amt)
{
	if(which >= 2)
		return;
	
	players[which].score += amt;
}

void shooty_players_awardlives(unsigned char which, unsigned char amt)
{
	if(which >= 2)
		return;
	
	players[which].extralives += amt;
	
	//Need to kick them out of INIT immediately, in case the players aren't ticking right now
	if(players[which].state == PLAYER_STATE_INIT && players[which].extralives > 0)
	{
		players[which].state = PLAYER_STATE_SPAWNING;
		players[which].stateframe = 0;
		players[which].extralives--;
	}
}
