//spr.c
//Actual sprite/plane rendering methods
//Called by the video system in hblank interrupts
//Input data should come from application program
//Bryan Topp <betopp@betopp.com> 2015

#include <string.h>
#include <stdio.h>

#include "textfont.h"
#include "vdac.h"
#include "spr.h"

//Used to be hardcoded palette indexes for test patterns
//Now just hardcoded colors for test patterns
#define CLR_IDX_Z 0x8000
#define CLR_IDX_R 0x801F
#define CLR_IDX_G 0x83E0
#define CLR_IDX_B 0xFC00
#define CLR_IDX_W 0xFFFF


//----------------------------------------
//Text plane
//----------------------------------------

__attribute__((aligned(16),section(".l1.bss.B"))) char spr_textplane[SPR_TEXTPLANE_ROWS][SPR_TEXTPLANE_COLS] = {{0}};
unsigned short spr_textplane_color = 0xFFFF;

void spr_textplane_clear()
{
	memset(spr_textplane, '\0', sizeof(spr_textplane));
	spr_textplane_color = 0xFFFF;
}

//----------------------------------------
//Sprites
//----------------------------------------

__attribute__((aligned(16),section(".l1.bss.B"))) spr_spritedef spr_spritetable[256];
unsigned char spr_sprite_enable = 0;

void spr_sprite_clear()
{
	for(unsigned short s = 0; s < 256; s++)
	{
		spr_spritetable[s].h = 0;
	}
}

//----------------------------------------
//Palette for sprites
//----------------------------------------

__attribute__((aligned(16))) static unsigned short spr_palette[256] = {0};

void spr_loadpalette_bgr(const unsigned int *bgrz8888)
{
	for(unsigned short entry = 0; entry < 255; entry++) //Last entry is not overwritten
	{
		//Unpack the 32-bit pixel
		unsigned short b = bgrz8888[entry] & 0xFF;
		unsigned short g = (bgrz8888[entry]>>8) & 0xFF;
		unsigned short r = (bgrz8888[entry]>>16) & 0xFF;
		
		//This was the formula used on the dev kit with its fancy-shmancy D1-style video codec
		//unsigned char Y  = (4096  + (65*r) + (129*g) + (25*b))>>8;
		//unsigned char Cb = (32768 - (37*r) - (74*g) + (112*b))>>8;
		//unsigned char Cr = (32768 + (112*r) - (94*g) - (18*b))>>8;
		//spr_palette[entry] = (Y<<24) | (Cr<<16) | (Y<<8) | Cb;
		
		//Pack into 15-bit RGB for the actual arcade RGB output
		spr_palette[entry] = 0x8000; //We REALLY shouldn't define a color with ~SYNC cleared.
		spr_palette[entry] |= ((b/8) << 10) & (0x7C00);
		spr_palette[entry] |= ((g/8) << 5)  & (0x03E0);
		spr_palette[entry] |= ((r/8) << 0)  & (0x001F);
	}
}


//----------------------------------------
//Background layer (under everything)
//----------------------------------------
const unsigned short *spr_background = NULL;
short spr_background_scroll_v = 0;

//----------------------------------------
//Test patterns
//----------------------------------------

unsigned short spr_testpattern = 0;

//Test pattern - dot grid
static void spr_render_pattern_dot(unsigned short *buffer_out, unsigned short y)
{
	if(y%16 == 0 || y%16 == 15)
	{
		for(unsigned short px = 0; px < SPR_ACTIVE_PX; px++)
			*(buffer_out++) = CLR_IDX_W;
	}
	else if(y % 16 == 7 || y % 16 == 8)
	{
		for(unsigned short px = 0; px < SPR_ACTIVE_PX; px++)
			*(buffer_out++) = (px%16==7||px%16==8||px%16==0||px%16==15)?CLR_IDX_W:CLR_IDX_Z;
	}
	else
	{
		for(unsigned short px = 0; px < SPR_ACTIVE_PX; px++)
			*(buffer_out++) = (px%16==0||px%16==15)?CLR_IDX_W:CLR_IDX_Z;
	}
}

//Test pattern - crosshatch
static void spr_render_pattern_cross(unsigned short *buffer_out, unsigned short y)
{	
	if(y%16 == 7)
	{
		for(unsigned short px = 0; px < SPR_ACTIVE_PX; px++)
			*(buffer_out++) = (px%16>4 && px%16 < 10)?CLR_IDX_W:CLR_IDX_Z;
	}
	else if(y%16 < 10 && y%16 > 4)
	{
		for(unsigned short px = 0; px < SPR_ACTIVE_PX; px++)
			*(buffer_out++) = (px%16==7)?CLR_IDX_W:CLR_IDX_Z;
	}
	else
	{
		for(unsigned short px = 0; px < SPR_ACTIVE_PX; px++)
			*(buffer_out++) = CLR_IDX_Z;
	}	
}

//Test pattern - colorbars + palette
static void spr_render_pattern_bars(unsigned short *buffer_out, unsigned short y)
{	
	if(y == 0 || y == 239) //Top/bottom borders (r/g/b)
	{
		for(unsigned short px = 0; px < 107; px++)
			*(buffer_out++) = CLR_IDX_R;
		for(unsigned short px = 107; px < 213; px++)
			*(buffer_out++) = CLR_IDX_G;
		for(unsigned short px = 213; px < 320; px++)
			*(buffer_out++) = CLR_IDX_B;
	}
	else if(y >= 32 && y < 208) //Ramps
	{
		*(buffer_out++) = CLR_IDX_R;
		
		for(unsigned short px = 1; px < 32; px++)
			*(buffer_out++) = 0xBDEF;
		
		if(y >= 48 && y < 72)
		{
			for(unsigned short px = 0; px < 256; px++) //Red ramp
				*(buffer_out++) = 0x8000 + (px/8);
		}
		else if(y >= 88 && y < 112)
		{
			for(unsigned short px = 0; px < 256; px++) //Green ramp
				*(buffer_out++) = 0x8000 + ((px/8)<<5);
		}
		else if(y >= 128 && y < 152)
		{
			for(unsigned short px = 0; px < 256; px++) //Blue ramp
				*(buffer_out++) = 0x8000 + ((px/8)<<10);
		}
		else if(y >= 168 && y < 192)
		{
			for(unsigned short px = 0; px < 256; px++) //White ramp
				*(buffer_out++) = 0x8000 + (px/8) + ((px/8)<<5) + ((px/8)<<10);
		}
		else
		{
			for(unsigned short px = 0; px < 256; px++) //Grey
				*(buffer_out++) = 0xBDEF;
		}
		
		for(unsigned short px = 0; px < 31; px++)
			*(buffer_out++) = 0xBDEF;
		
		*(buffer_out++) = CLR_IDX_B;
	}
	else //Blank lines - still have red/blue border
	{
		*(buffer_out++) = CLR_IDX_R;
		
		for(unsigned short px = 1; px < 319; px++)
			*(buffer_out++) = CLR_IDX_Z;
		
		*(buffer_out++) = CLR_IDX_B;
	}
}


//----------------------------------------
//Main interface
//----------------------------------------

void spr_init()
{	
	//Set palette to all black except the last
	for(unsigned short p = 0; p < 256; p++)
		spr_palette[p] = 0x8000;
	
	spr_palette[255] = 0xFFFF; //Until we load a real palette, make sure at least white is here
	
	spr_textplane_clear();
	
	spr_sprite_enable = 0;
	
	spr_spritetable[0].data = NULL;
	
	spr_background = NULL;
	spr_background_scroll_v = 0;
	
	printf("Init line rendering engine...\n");
	
	strcpy((char*)spr_textplane[2], "  ADSP-BF533 Game BIOS (c) 2016 betopp  ");
	strcpy((char*)spr_textplane[3], "  VDAC DMA and rendering engine online  ");
}

void spr_render_line(unsigned short *buffer_out, unsigned char y)
{		
	//Test patterns take over line rendering entirely
	if(spr_testpattern == SPR_TESTPAT_DOT)
	{		
		spr_render_pattern_dot(buffer_out, y);
		return;
	}
	
	if(spr_testpattern == SPR_TESTPAT_CROSS)
	{
		spr_render_pattern_cross(buffer_out, y);
		return;
	}
	
	if(spr_testpattern == SPR_TESTPAT_BARS)
	{
		spr_render_pattern_bars(buffer_out, y);
		return;
	}
	//Otherwise, do normal rendering
	
	
	//Fill with background if we have one, or solid black otherwise
	if(spr_background)
	{
		unsigned short bgwidth = ((const unsigned short*)spr_background)[4];
		unsigned short bgheight = ((const unsigned short*)spr_background)[6];
		
		unsigned short bgrow = (y+(spr_background_scroll_v%bgheight)+bgheight) % bgheight;
		
		memcpy(buffer_out, spr_background + (bgrow * bgwidth) + 8, SPR_ACTIVE_PX*2);
	}
	else
	{
		for(unsigned short px = 0; px < SPR_ACTIVE_PX; px++)
			buffer_out[px] = 0x8000;
	}

	//Render sprites
	if(spr_sprite_enable)
	{	
		//Can be kept across iterations to chain sprites, like Neo-Geo
		unsigned short y_in_sprite = 0;
		short startx = 0;

		const spr_spritedef *sptr = spr_spritetable-1; //Pointer to our place in the sprite table
		
		for(unsigned char spr = 0; spr < 128; spr++)
		{	
			sptr++;
			
			//Find out where we are vertically in relation to the sprite.
			//Keep the last vertical positioning if "chain" is enabled.
			y_in_sprite = (sptr->y != SPR_CHAIN_COORDINATE)?(y - sptr->y):(y_in_sprite);
			
			//Find out the horizontal location of the sprite
			//Chain similarly, except advance to the right when chaining
			//Note: don't early-out before this, to avoid breaking chains!
			startx = (sptr->x != SPR_CHAIN_COORDINATE)?(sptr->x):(startx+16);
		
			//If we're not in Y range of the sprite, ignore it
			//(Note that y_in_sprite is unsigned... a negative result is thus treated as huge.)
			if(y_in_sprite >= sptr->h)
				continue;
			
			//Compute alignment of sprite data and line data
			//(i.e. the correct line of the sprite, shifted based on its X position)
			const unsigned char * const sprite_line_aligned = sptr->data + (y_in_sprite*16) - startx;
			
			//Render the pixels in range
			short clip_st = (startx>0)?startx:0;
			short clip_ed = (SPR_ACTIVE_PX<startx+16)?(SPR_ACTIVE_PX):(startx+16);			
			
			for(short p = clip_st; p < clip_ed; p++)
				if(sprite_line_aligned[p])
					buffer_out[p] = spr_palette[sprite_line_aligned[p]];
		}
	}
			
	//Render text plane over sprites.
	{
		const char * const textplane_text = spr_textplane[y/8];
		const unsigned char * const textfont_bits_shifted = textfont_bits + (y%8);
		
		short x = 0;
		
		for(unsigned char col = 0; col < (SPR_ACTIVE_PX/8); col++)
		{
			//Find out which character is next and fetch a row of it		
			const unsigned char textplane_octopx = textfont_bits_shifted[textplane_text[col]*8];

			if(!(textplane_octopx & 1)) 
				buffer_out[x+0] = spr_textplane_color;
			if(!(textplane_octopx & 2)) 
				buffer_out[x+1] = spr_textplane_color;
			if(!(textplane_octopx & 4)) 
				buffer_out[x+2] = spr_textplane_color;
			if(!(textplane_octopx & 8)) 
				buffer_out[x+3] = spr_textplane_color;
			if(!(textplane_octopx & 16)) 
				buffer_out[x+4] = spr_textplane_color;
			if(!(textplane_octopx & 32)) 
				buffer_out[x+5] = spr_textplane_color;
			if(!(textplane_octopx & 64)) 
				buffer_out[x+6] = spr_textplane_color;
			if(!(textplane_octopx & 128)) 
				buffer_out[x+7] = spr_textplane_color;
			
			x += 8;
		}	
	}
	
	
}

