//testmenu.c
//Hardware test/service menu
//Bryan Topp <betopp@betopp.com> 2015


#include "testmenu.h"
#include "spr.h"
#include "vdac.h"
#include "pads.h"
#include "dram.h"
#include "dramfs.h"
#include "adac.h"
#include "fram.h"

#include <string.h>
#include <stdio.h>

static const char *testmenu_header[] = 
{
	"  BF533 Game System test menu           ",
	"  ------------------------------------  ", 
	NULL
};

static const char *testmenu_footer[] =
{
	"  ------------------------------------  ", 
	"  Move cursor using up/down switches.   ",
	"  Select item at cursor with B1 switch. ",
	"  Return to prior menu with B2 switch.  ",
	NULL 
};
	
#define MENU_BODY_ROW_START 6
#define MENU_BODY_ROW_COUNT 15

#define MENU_BODY_INDENT 2

//"Generic" menus driven by this data
//Each item has a name, and - if it's selectable - a code/data pointer set
typedef struct testmenu_generic_data_s
{
	const char *name;
	void (*codeptr)(const void *dataptr);
	const void *dataptr;
} testmenu_generic_data;

static void testmenu_wait_noinput()
{
	while(1)
	{
		pads_poll();
		vdac_fieldwait();
		if(pads_firstplayer == 0)
			break;
	}
}

static void testmenu_wait_anyinput()
{
	while(1)
	{
		pads_poll();
		vdac_fieldwait();
		if(pads_firstplayer != 0)
			break;
	}	
}

static void testmenu_predraw()
{
	//Clear the text plane and paint in the headers
	spr_textplane_clear();
	
	for(int l = 0; testmenu_header[l] != NULL; l++)
		strcpy(spr_textplane[l+2], testmenu_header[l]);
	
	for(int l = 0; testmenu_footer[l] != NULL; l++)
		strcpy(spr_textplane[l+24], testmenu_footer[l]);	
}

static void testmenu_generic_redraw(const testmenu_generic_data *data, unsigned short selected_option)
{
	testmenu_predraw();
	
	unsigned short line = MENU_BODY_ROW_START;
	
	//Draw each of the options
	for(int i = 0; data[i].name != NULL; i++)
	{
		strcpy(spr_textplane[line]+MENU_BODY_INDENT+1, data[i].name);
		
		line++;
		if(line >= MENU_BODY_ROW_START + MENU_BODY_ROW_COUNT)
			break;
	}
	
	//Paint the cursor on the selected option
	if(data[selected_option].name != NULL)
		spr_textplane[selected_option + MENU_BODY_ROW_START][MENU_BODY_INDENT] = '$';
}


static unsigned char testmenu_generic_depth = 0;
static unsigned char testmenu_done = 0; //Globally causes all testmenu_generic to return

static void testmenu_setdone(const void *unused)
{
	testmenu_done = 1;
}

static void testmenu_generic(const void *dataptr)
{	
	const testmenu_generic_data *data = (const testmenu_generic_data*)dataptr;
	
	//Start the cursor on the first valid option
	unsigned short selected_option = 0;
	while(data[selected_option].name != NULL && data[selected_option].codeptr == NULL)
		selected_option++;
		
	//Draw the menu
	testmenu_generic_redraw(data, selected_option);

	while(!testmenu_done) //Interaction loop, can be cancelled by global "done"
	{				
		//Paint the cursor on the selected option
		if(data[selected_option].name != NULL)
			spr_textplane[selected_option + MENU_BODY_ROW_START][MENU_BODY_INDENT] = '$';
				
		//Wait for input	
		testmenu_wait_noinput();
		testmenu_wait_anyinput();
		
		//Handle that input
		printf("Test menu got input 0x%.2X\n", pads_firstplayer);
		
		if(pads_firstplayer & PADS_BIT_DOWN)
		{
			//Don't allow input if we have no valid options
			if(data[selected_option].name != NULL)
			{
				//Kick the option down one, and then advance past any non-selectable options
				unsigned short new_option = selected_option + 1;
				while(data[new_option].name != NULL && data[new_option].codeptr == NULL)
					new_option++;
				
				//If we found a new option, move to it
				if(data[new_option].name != NULL)
				{
					spr_textplane[selected_option + MENU_BODY_ROW_START][MENU_BODY_INDENT] = 0;
					
					selected_option = new_option;
					spr_textplane[selected_option + MENU_BODY_ROW_START][MENU_BODY_INDENT] = '$';
				}
			}
		}
		else if(pads_firstplayer & PADS_BIT_UP)
		{
			if(data[selected_option].name != NULL)
			{
				//Similar to "down" case, but in reverse
				unsigned short new_option = selected_option - 1;
				while(new_option != 0xFFFF && data[new_option].codeptr == NULL)
					new_option--;
				
				//If we found a new option, move to it
				if(new_option != 0xFFFF)
				{
					spr_textplane[selected_option + MENU_BODY_ROW_START][MENU_BODY_INDENT] = 0;
					
					selected_option = new_option;
					spr_textplane[selected_option + MENU_BODY_ROW_START][MENU_BODY_INDENT] = '$';
				}
			}
		}
		else if(pads_firstplayer & PADS_BIT_B1)
		{
			if(data[selected_option].codeptr != NULL)
			{
				printf("Selecting %s\n", data[selected_option].name);			
				
				//Call the function
				testmenu_generic_depth++;
				data[selected_option].codeptr(data[selected_option].dataptr);
				testmenu_generic_depth--;
				
				//When it returns, redraw
				testmenu_generic_redraw(data, selected_option);
			}
		}
		else if(pads_firstplayer & PADS_BIT_B2)
		{
			//Don't let us back out from the main menu
			if(testmenu_generic_depth > 0)
				return;
			else
				printf("Not backing out of main...\n");
		}
	}
	
	
}

//Activates a VDAC test pattern until input is pressed.
static void testmenu_spr_pattern_activate(const void* patternp)
{
	//not actually a pointer right now
	int pattern = (unsigned long)patternp;
	
	spr_testpattern = pattern;
	
	testmenu_wait_noinput();
	testmenu_wait_anyinput();
	
	spr_testpattern = 0;
	
	return;
}

static void testmenu_switchtest(const void *unused)
{
	testmenu_predraw();
	
	static const char *menutext[] = {
		" -Switch Test- P1 P2      - Open",
		"                          X Closed",
		"         Left:",
		"        Right:",
		"           Up:",
		"         Down:",
		"           B1:         Coin 1:",
		"           B2:         Coin 2:",
		"           B3:        Service:",
		"        Start:           Test:",
		"",
		" Press P1 B1+Start to return.",
		NULL};
	
	for(short i = 0; menutext[i] != NULL; i++)
		strcpy(&(spr_textplane[MENU_BODY_ROW_START+i][MENU_BODY_INDENT]), menutext[i]);
	
	const short p1_column = MENU_BODY_INDENT + strlen(menutext[2])+1;
	const short p2_column = MENU_BODY_INDENT + strlen(menutext[2])+4;
		
	while(1)
	{
		pads_poll();
		vdac_fieldwait();
		
		for(short bit=0; bit < 8; bit++)
		{
			if(pads_firstplayer & (1<<bit))
				spr_textplane[MENU_BODY_ROW_START+2+bit][p1_column] = 'X';
			else
				spr_textplane[MENU_BODY_ROW_START+2+bit][p1_column] = '-';
			
			if(pads_secondplayer & (1<<bit))
				spr_textplane[MENU_BODY_ROW_START+2+bit][p2_column] = 'X';
			else
				spr_textplane[MENU_BODY_ROW_START+2+bit][p2_column] = '-';
		}
		
		spr_textplane[MENU_BODY_ROW_START+6][33] = (pads_sysbuttons&PADS_SYS_BIT_COIN1)?'X':'-';
		spr_textplane[MENU_BODY_ROW_START+7][33] = (pads_sysbuttons&PADS_SYS_BIT_COIN2)?'X':'-';
		spr_textplane[MENU_BODY_ROW_START+8][33] = (pads_sysbuttons&PADS_SYS_BIT_SERVICE)?'X':'-';
		spr_textplane[MENU_BODY_ROW_START+9][33] = (pads_sysbuttons&PADS_SYS_BIT_TEST)?'X':'-';
		
		if( (pads_firstplayer & PADS_BIT_START) && (pads_firstplayer & PADS_BIT_B1) )
			return;
	}
}

static void testmenu_dramfs(const void *unused)
{
	testmenu_predraw();
	
	strcpy(&(spr_textplane[MENU_BODY_ROW_START][MENU_BODY_INDENT+1]), "-DRAMFS information-");
	
	if(!dramfs_isvalid())
	{
		strcpy(&(spr_textplane[MENU_BODY_ROW_START+2][MENU_BODY_INDENT+1]), "DRAMFS not loaded.");
		strcpy(&(spr_textplane[MENU_BODY_ROW_START+3][MENU_BODY_INDENT+1]), "Testing DRAM clobbers DRAMFS.");
		strcpy(&(spr_textplane[MENU_BODY_ROW_START+4][MENU_BODY_INDENT+1]), "Reset machine to re-load.");
	}
	else
	{
	
		strcpy(&(spr_textplane[MENU_BODY_ROW_START+2][MENU_BODY_INDENT+1]), "DRAMFS loaded successfully.");
		snprintf(&(spr_textplane[MENU_BODY_ROW_START+3][MENU_BODY_INDENT+1]), SPR_TEXTPLANE_COLS - (MENU_BODY_INDENT+1),
			"Blob in memory contains %u files.", dramfs_numfiles());
		
		//Browse loop
		/*while(1)
		{
		
			for(unsigned int i = 0; i < 16; i++)
			{
				
			}
		}*/
	}

	strcpy(&(spr_textplane[MENU_BODY_ROW_START+MENU_BODY_ROW_COUNT-1][MENU_BODY_INDENT+1]), "Press any button to return.");
	testmenu_wait_noinput();
	testmenu_wait_anyinput();
}


static unsigned int lfsr32_advance(unsigned int old)
{
	if(old & 1)
		return (old>>1) ^ (0xA3000000 /*Galois taps*/);
	else
		return old>>1;	
}

static unsigned int lfsr24_advance(unsigned int old)
{
	old &= 0x00FFFFFF;
	
	if(old & 1)
		return (old>>1) ^ (0x00D80000 /*Galois taps*/);
	else
		return old>>1;
}


#ifndef JSYS_PC_MOCKUP
	#include <sys/platform.h> //For turning DCPLBs off and on
#endif

static void testmenu_dramcheck(const void *unused)
{
	dramfs_clobbered();
	
	testmenu_predraw();
	
	strcpy(&(spr_textplane[MENU_BODY_ROW_START][MENU_BODY_INDENT+1]), "-Read/write test-");
	
	#ifndef JSYS_PC_MOCKUP
		printf("Turning off DCPLBs...\n");
		ssync();
		*pDMEM_CONTROL &= ~ENDCPLB;
		ssync();
	#endif
	
	printf("DRAM check write...\n");
	strcpy(&(spr_textplane[MENU_BODY_ROW_START+2][MENU_BODY_INDENT+1]), "Writing 16Mx16 words...");
	
	//Write data into DRAM
	////
	volatile unsigned short *dram = DRAM_ORIGIN;
	unsigned int addr_prng = 1234u;
	unsigned int data_prng = 4321u;
		
	do {		
		dram[addr_prng] = data_prng;
		
		addr_prng = lfsr24_advance(addr_prng);
		data_prng = lfsr32_advance(data_prng);
	} while(addr_prng != 1234u);	
	////
	
	
	printf("DRAM check read...\n");
	strcpy(&(spr_textplane[MENU_BODY_ROW_START+3][MENU_BODY_INDENT+1]), "Reading back...");
	
	//Read it back and verify
	////
	addr_prng = 1234u;
	data_prng = 4321u;
	do {	
		unsigned short existing_data = dram[addr_prng];
		unsigned short expected_data = (unsigned short)data_prng;
		
		if(existing_data != expected_data)
		{
			
			fprintf(stderr, "DRAM ERROR DETECTED!\n");
			fprintf(stderr, "Address:   %p\n", &(dram[addr_prng]));
			fprintf(stderr, "Read data: %X\n", existing_data);
			fprintf(stderr, "Expected:  %X\n", expected_data);
			
			strcpy(&(spr_textplane[MENU_BODY_ROW_START+4][MENU_BODY_INDENT+1]), "!! DRAM ERROR !!");
			sprintf(&(spr_textplane[MENU_BODY_ROW_START+5][MENU_BODY_INDENT+1]), "Addr: %p", &(dram[addr_prng]));
			sprintf(&(spr_textplane[MENU_BODY_ROW_START+6][MENU_BODY_INDENT+1]), "Read: %X", existing_data);
			sprintf(&(spr_textplane[MENU_BODY_ROW_START+7][MENU_BODY_INDENT+1]), "Want: %X", expected_data);
			
			break;
		}
		
		addr_prng = lfsr24_advance(addr_prng);
		data_prng = lfsr32_advance(data_prng);
	} while(addr_prng != 1234u);
	////
		
	if(addr_prng == 1234u) //Made it all the way back around
	{
		printf("Success.\n");
		strcpy(&(spr_textplane[MENU_BODY_ROW_START+4][MENU_BODY_INDENT+1]), "No errors found.");
	}
	
	#ifndef JSYS_PC_MOCKUP
		printf("Turning DCPLBs back on...\n");
		ssync();
		*pDMEM_CONTROL |= ENDCPLB;
		ssync();	
	#endif
	
	strcpy(&(spr_textplane[MENU_BODY_ROW_START+MENU_BODY_ROW_COUNT-1][MENU_BODY_INDENT+1]), "Press any button to return.");
	testmenu_wait_noinput();
	testmenu_wait_anyinput();
}


static void testmenu_audio_set(const void *audiosetting)
{
	unsigned char newsetting = (unsigned long)audiosetting;
	adac_testmode = newsetting;
	
	strcpy(spr_textplane[18] + 2, "Playing... press any input to abort.");
	
	testmenu_wait_noinput();
	testmenu_wait_anyinput();
	
	adac_testmode = 0;
}

static const testmenu_generic_data testmenu_generic_aud[] =
{
	{"-Audio DAC test-", NULL, NULL},
	{"", NULL, NULL},
	{"1500Hz triangle, 50% magnitude", NULL, NULL},
	{"", NULL, NULL},
	{"Left channel", testmenu_audio_set, (void*)1},
	{"Right channel", testmenu_audio_set, (void*)2},
	{"Both channels", testmenu_audio_set, (void*)3},
	{NULL, NULL, NULL}
};


static void testmenu_lenna(const void *dummy)
{
	if(dramfs_isvalid())
	{
		spr_background = dramfs_getptr(dramfs_findfile("bg/lenna.solid16"));
		spr_textplane_clear();
	}
	else
	{
		strcpy(spr_textplane[18] + 2, "DRAMFS not valid. Press a button.");
	}
	
	testmenu_wait_noinput();
	testmenu_wait_anyinput();
	
	spr_background = NULL;
}

static const testmenu_generic_data testmenu_generic_dram[] =
{
	{"-DRAM diagnostics-", NULL, NULL},
	{"", NULL, NULL},
	{"DRAMFS information",   testmenu_dramfs, NULL},
	{"Read/write test",      testmenu_dramcheck, NULL},
	{"Lenna",                testmenu_lenna, NULL},
	{NULL, NULL, NULL}
};

static const testmenu_generic_data testmenu_generic_spr[] =
{
	{"-Video line test-", NULL, NULL},
	{"", NULL, NULL},
	{"Press any button to end tests.",NULL, NULL},
	{"",                           NULL, NULL},
	{"Dot grid",             testmenu_spr_pattern_activate, (void*)SPR_TESTPAT_DOT},
	{"Color ramps",           testmenu_spr_pattern_activate, (void*)SPR_TESTPAT_BARS},
	{"Crosshatch",           testmenu_spr_pattern_activate, (void*)SPR_TESTPAT_CROSS},
	{NULL, NULL, NULL}
};

static void testmenu_iocon_reload(const void *dummy)
{
	extern void padloader();
	padloader();
}

static void testmenu_fram_reset_confirm(const void *dummy)
{
	if(dummy != NULL)
	{
		fram_nuke();
		sprintf(spr_textplane[18] + 2, "FRAM cleared. Press any switch.");
		testmenu_wait_noinput();
		testmenu_wait_anyinput();
	}
}

static const testmenu_generic_data testmenu_generic_fram_reset[] = 
{
	{"-Reset FRAM-", NULL, NULL},
	{"", NULL, NULL},
	{"This will clear all bookkeeping", NULL, NULL},
	{"and high-score data.", NULL, NULL},
	{"", NULL, NULL},
	{"Are you sure?", NULL, NULL},
	{"No", testmenu_fram_reset_confirm, 0},
	{"Yes", testmenu_fram_reset_confirm, (void*)1},
	{NULL, NULL, NULL}
};

static void testmenu_bookkeeping(const void *dummy)
{
	sprintf(spr_textplane[13] + 2, "First slot:  %10u", fram_shadow.coins_deposited_1);
	sprintf(spr_textplane[14] + 2, "Second slot: %10u", fram_shadow.coins_deposited_2);
	sprintf(spr_textplane[16] + 2, "Total:       %10u", fram_shadow.coins_deposited_1 + fram_shadow.coins_deposited_2);

	sprintf(spr_textplane[18] + 2, "Press any switch.");
	
	testmenu_wait_noinput();
	testmenu_wait_anyinput();
}

static const testmenu_generic_data testmenu_generic_fram[] = 
{
	{"-FRAM data-", NULL, NULL},
	{"",  NULL, NULL},
	{"Coin counts",      testmenu_bookkeeping, NULL},
	{"Reset FRAM",            testmenu_generic, testmenu_generic_fram_reset},
	{NULL, NULL, NULL}
};

static const testmenu_generic_data testmenu_generic_version[] = 
{
	{"-Version / copyright-", NULL, NULL},
	{"",                    NULL, NULL},
	{"BF533 Game System (c) 2016 betopp",   NULL, NULL},
	{"",                    NULL, NULL},
	{"Git version:",        NULL, NULL},
	{" "JSYS_GIT_VERSION,   NULL, NULL},
	{"",                    NULL, NULL},
	{"Build date:",         NULL, NULL},
	{" "JSYS_BUILD_DATE,    NULL, NULL},
	{"",                    NULL, NULL},
	{"Builder:",            NULL, NULL},
	{" "JSYS_BUILDER,       NULL, NULL},
	{"",                    NULL, NULL},
	#ifdef JSYS_PC_MOCKUP
	{"PC MOCKUP BUILD",     NULL, NULL},
	#else
	{"BF533 TARGET BUILD",     NULL, NULL},
	#endif
	{NULL, NULL, NULL}
};
	

static const testmenu_generic_data testmenu_generic_main[] = 
{
	{"-Main Menu-",           NULL, NULL},
	{"",                      NULL, NULL},
	{"Video line test",       testmenu_generic, testmenu_generic_spr},
	{"Audio DAC test",        testmenu_generic, testmenu_generic_aud},
	{"DRAM diagnostics",      testmenu_generic, testmenu_generic_dram},
	{"Switch test",           testmenu_switchtest, NULL},
	{"Force IOCON reflash",   testmenu_iocon_reload, NULL},
	{"FRAM data",             testmenu_generic, testmenu_generic_fram},
	{"",                      NULL, NULL},
	{"Version / copyright",   testmenu_generic, testmenu_generic_version},
	{"",	                  NULL, NULL},
	{"Exit",	          testmenu_setdone, NULL},
	{NULL, NULL, NULL}
};
	
void testmenu()
{	
	printf("Beginning test menu.\n");
	
	testmenu_done = 0;
	testmenu_generic_depth = 0;
	testmenu_generic(testmenu_generic_main);
	
	printf("Test menu exited.\n");
}
