//uart.c
//Helpers for newlib's implementation of printf/etc
//Sets up our use of the Blackfin UART for stdin/stdout
//Bryan Topp <betopp@betopp.com> 2015
//Follows the ADI guide here: http://blackfin.uclinux.org/doku.php?id=toolchain:bare_metal:newlib

#ifndef JSYS_PC_MOCKUP

#include <sys/platform.h>
#include <stdio.h>
#include <errno.h>
#include <sys/stat.h>

//Initializes the Blackfin's UART
#define UART_BAUD 115200u

void uart_init()
{
	//UART clock enable
	*pUART_GCTL = UCEN;
	
	//Store divisor
	unsigned short divisor = (unsigned short)(98304000u/(16u*UART_BAUD) + 1u);
	*pUART_LCR = DLAB;
	*pUART_DLL = divisor & 0xFF;    //Divisor Low
	*pUART_DLH = (divisor>>8) & 0xFF; //Divisor High
	
	//8N1
	*pUART_LCR = WLS(8);
	
	
	//Raw write an introductory string to test hardware.
	const char *s = "BF533 UART online.\n\rIf you can read this, the metal is working.\n\rProceeding to try stdout...\n\r";
	while(*s)
	{
		while (!(*pUART_LSR & THRE)) { };
		*pUART_THR = *s;
		s++;
	}
	
	//Test STDIO
	printf("Hello from newlib stdout!\n");
	fprintf(stderr, "Hello from newlib stderr!\n");
	
}


//UART driving functions
//----------------------
inline static void uart_wait_txrdy()
{
	//Wait for the transmit buffer to be nonfull
	while (!(*pUART_LSR & THRE)) {};
}

inline static void uart_wait_rxrdy()
{
	//Wait for the receive buffer to be nonempty
	while (!(*pUART_LSR & DR)) {};
}

inline static void uart_putc(char c)
{
	uart_wait_txrdy();
	//Place one character in the transmit buffer
	*pUART_THR = c;
	
	if(c == '\n') //Hack
	{
		uart_wait_txrdy();
		*pUART_THR = '\r';
	}
}

inline static char uart_getc()
{
	uart_wait_rxrdy();
	//Grab one character from the receive buffer
	return *pUART_RBR;
}

inline static void uart_wait_txdone()
{
	//Wait for transmit buffer to be empty
	while (!(*pUART_LSR & TEMT)) {};
		
	//Wait for speculative state to settle
	ssync();
}

inline static int uart_gets(char *str, int max)
{
	//Grab characters until end-of-line or max
	for(int i = 0; i < max; i++)
	{
		str[i] = uart_getc();
		if (str[i] == '\n')
		{
			return i+1;
		}
	}
	return max;
}


//Hookup to newlib's file operations
//----------------------------------

//We don't support opening new files.
int _open(const char *name, int flags, int mode)
{	
	errno = ENOSYS;
	return -1;
}

//We don't support closing files.
int _close(int file)
{
	errno = EBADF;
	return -1;
}

//Writes go out the UART
int _write(int file, char *buf, int nbytes)
{
	//Spit out coloration first
	uart_putc(0x1B /* \e */);
	uart_putc('[');
	uart_putc('3');
	uart_putc((file>1)?'1':'9'); //stderr is red, stdout is default
	uart_putc(';');
	uart_putc('4');
	uart_putc('9');
	uart_putc('m');
	
	for (int i = 0; i < nbytes; i++)
		uart_putc( *buf++ );

	//reset coloring
	const char *rstseq = "\x1b[39;49m";
	for(int i = 0; i < 8; i++)
		uart_putc(rstseq[i]);
	
	uart_wait_txdone();

	return nbytes;
}

//Reads do UART reading
int _read(int file, char *buf, int nbytes)
{
	return uart_gets( buf, nbytes );
}

//Stat'ing a file just says that it's a character device.
int _fstat(int file, struct stat *st)
{
	st->st_mode = S_IFCHR;
	return 0;
}

//Seek does nothing.
int _lseek(int file, int offset, int whence)
{
	return 0;
}

//In fact, all we support are TTYs
int _isatty(int file)
{
	return 1;
}

#else
//Mockup follows
#include <stdio.h>
//No mockup code needed here. We have stdio support by default... I'd hope.
void uart_init() { printf("Dummy UART. stdio already exists.\n"); }
//End of mockup code
#endif //JSYS_PC_MOCKUP switch