//vdac.c
//Initializes video output, sets up timed calls of rendering function
//Bryan Topp <betopp@betopp.com> 2015

//The video DAC calls out to the sprite renderer or the audio system to render things
//(Sprites mostly - audio gets slotted in during V back porch)
#include "spr.h"
#include "adac.h"

#ifndef JSYS_PC_MOCKUP
//--------------------

#ifndef __ADSPBF533__
	#error This is for the Blackfin BF533.
#endif

#include <stdio.h>
#include <sys/platform.h>

#define HPULSE_START 329
#define HPULSE_END 358
#define LINE_PERIOD 390

#define FRAME_PERIOD 262

//DMA-Interrupt setup to render and output lines
//----------------------------------------------
unsigned short vdac_nextline = 0; //in frame, not in buffers
volatile unsigned short vdac_fieldcount = 0;
__attribute__((section(".l1.bss.B"),aligned(16))) unsigned short vdac_linebuffers[LINE_PERIOD*2];

__attribute__((interrupt_handler)) void vdac_dma_isr();
static inline void vdac_enable_interrupt()
{
	//Unmask DMA interrupt in core interrupt controller
	*pIMASK |= 1<<8;
	
	//Unmask DMA interrupt in system interrupt controller
	*pSIC_IMASK |= 1<<8;
	
	//Set interrupt vector
	*pEVT8 = &vdac_dma_isr;	
}

static inline void vdac_config_dma()
{
	*pDMA0_START_ADDR = vdac_linebuffers;
	
	//16-bit, turn on autobuffer, data interrupts, interrupts at each line, 2D DMA
	//See BF533 HW reference, P. 9-12
	*pDMA0_CONFIG = WDSIZE_16 | DMA2D | DI_EN | DI_SEL | FLOW_AUTO;
	
	*pDMA0_X_COUNT = LINE_PERIOD; //390 samples of RGB5551 at 6.144MHz
	*pDMA0_X_MODIFY = 2; //16-bit transfers...
	
	*pDMA0_Y_COUNT = 2;
	*pDMA0_Y_MODIFY = 2;
}

static inline void vdac_config_ppi()
{
	//Don't enable yet - wait for vdac_enable_ppi_dma
	*pPPI_CONTROL = PORT_DIR | DLEN_16;
}

static inline void vdac_enable_ppi_dma()
{
	//BF533 HW manual, p9-16:
	//To avoid unexpected results, take care to enable the DMA channel before enabling the peripheral
	
	*pDMA0_CONFIG |= 1;
	*pPPI_CONTROL |= 1;
}


//Interrupt handling - building the actual video lines
//----------------------------------------------------	

//Fills a line in the vertical active area
static inline void vdac_line_active(unsigned short *buffer_out, int l)
{
	//Active area
	spr_render_line(buffer_out, l);
	
	for(int xi=320; xi<HPULSE_START; xi++) //Front porch
		buffer_out[xi] = 0x8000;
	
	for(int xi=HPULSE_START; xi<HPULSE_END; xi++) //Pulse
		buffer_out[xi] = 0x0000;
	
	for(int xi=HPULSE_END; xi<LINE_PERIOD; xi++) //Back porch
		buffer_out[xi] = 0x8000;
}

//Fills a line in the vertical porch
static inline void vdac_line_porch(unsigned short *buffer_out)
{	
	for(int xi=0; xi<320;xi++) //Active area (i.e. black)
		buffer_out[xi] = 0x8000;
	
	for(int xi=320; xi<HPULSE_START; xi++) //Front porch
		buffer_out[xi] = 0x8000;
	
	for(int xi=HPULSE_START; xi<HPULSE_END; xi++) //Pulse
		buffer_out[xi] = 0x0000;
	
	for(int xi=HPULSE_END; xi<LINE_PERIOD; xi++) //Back porch
		buffer_out[xi] = 0x8000;
}

//Fills a line in the vertical sync pulse
static inline void vdac_line_pulse(unsigned short *buffer_out)
{
	//Identical to the vertical porch, but the polarity of the sync line is swapped (i.e. composite sync is HSync^VSync)
	
	for(int xi=0; xi<320;xi++) //Active area (i.e. black)
		buffer_out[xi] = 0x0000;
	
	for(int xi=320; xi<HPULSE_START; xi++) //Front porch
		buffer_out[xi] = 0x0000;
	
	for(int xi=HPULSE_START; xi<HPULSE_END; xi++) //Pulse
		buffer_out[xi] = 0x8000;
	
	for(int xi=HPULSE_END; xi<LINE_PERIOD; xi++) //Back porch
		buffer_out[xi] = 0x0000;
}

unsigned int vdac_perfstat;

//Draws the next line - including video sync/etc
//Calls an inner function to just draw active line data when needed
static inline void vdac_line_fill(unsigned short *buffer_out)
{
	//Grab the number of the line we're going to render, and increment right here.
	//(Prevent missing an increment and falling behind when we're dropping stuff)
	unsigned short l = vdac_nextline;
	vdac_nextline++;
	if(vdac_nextline >= FRAME_PERIOD)
	{
		vdac_nextline = 0;
	}
	
	if(l < 240)
	{
		vdac_line_active(buffer_out, l);
	}
	else if(l < 243)
	{
		if(l == 240) //Increment the field count just as we end active scan
			vdac_fieldcount++;
		
		vdac_line_porch(buffer_out);
	}
	else if(l < 246)
	{
		vdac_line_pulse(buffer_out);
	}
	else if(l < 248)
	{
		vdac_line_porch(buffer_out);
	}
	else
	{
		//After we've rendered two lines in our back porch, we don't really need to keep rendering
		//(We can output those each 8 times to make the porch.)
		//Instead, pump audio during this time
		adac_pump();
	}
}

//Interrupt handler for DMA completion
__attribute__((interrupt_handler)) void vdac_dma_isr()
{	
	//Get the count of remaining lines
	unsigned short l = *pDMA0_CURR_Y_COUNT;
	
	vdac_perfstat += *pDMA0_CURR_X_COUNT;
	
	//Build the next line based on which we just finished
	if(l%2)
		vdac_line_fill(vdac_linebuffers);
	else
		vdac_line_fill(vdac_linebuffers+LINE_PERIOD);

	vdac_perfstat -= *pDMA0_CURR_X_COUNT;
	
	//Handled
	*pDMA0_IRQ_STATUS |= 1;
}




//Outline - initialization code flow
//----------------------------------

void vdac_init()
{
	printf("VDAC init:\n");
	
	//Perform init sequence	
	printf("Activating line interrupts...\n");
	vdac_enable_interrupt();
	
	printf("Configuring DMA chain...\n");
	vdac_config_dma();
	
	printf("Configuring PPI output...\n");
	vdac_config_ppi();
	
	printf("Flipping the switch...\n");
	vdac_enable_ppi_dma();
			
	printf("VDAC init complete, check display.\n");
}

//Helper - wait for VSync
void vdac_fieldwait()
{
	unsigned short oldfield = vdac_fieldcount;
	do {
		asm("IDLE;");
	} while(vdac_fieldcount == oldfield);
	
	if(vdac_fieldcount % 64 == 0)
	{
		//printf("VDAC perf: %u\n", vdac_perfstat);
		vdac_perfstat = 0;
	}
}


#else
//Mockup follows


#include <SDL2/SDL.h>
#include <assert.h>

SDL_Window *sdlWin;
SDL_Renderer *sdlRen;	
SDL_Texture *rendertex;

volatile unsigned int vdac_fieldcount = 0;
unsigned int vdac_perfstat = 0;

int vdac_renderthread(void *data)
{
	SDL_Init(SDL_INIT_VIDEO);
	//Make window/renderer
	SDL_SetHint(SDL_HINT_RENDER_VSYNC, "1");
	int err = SDL_CreateWindowAndRenderer(640, 480, /*flags*/ 0, &sdlWin, &sdlRen);
	assert(!err);
	assert(sdlWin);
	assert(sdlRen);
	
	//Make texture for drawing
	rendertex = SDL_CreateTexture(sdlRen, SDL_PIXELFORMAT_ABGR1555, SDL_TEXTUREACCESS_STREAMING, 320, 240);
	SDL_RenderSetLogicalSize(sdlRen, 320, 240);	

	while(1)
	{
		SDL_RenderClear(sdlRen);
	
		unsigned short *px;
		int pxpitch;
		SDL_LockTexture(rendertex, NULL, (void**)&px, &pxpitch);
	
		for(int i = 0; i < 240; i++)
		{
			//Grab slice of texture and render sprite line into it
			spr_render_line(px, i);
			px += pxpitch/2; //pxpitch is in bytes, px is in shorts (hooray for pointer arithmetic)
		}
		
		for(int i = 0; i < 8; i++)
		{
			adac_pump();
		}
		
		vdac_fieldcount++;
		
		SDL_UnlockTexture(rendertex);
		SDL_RenderCopy(sdlRen, rendertex, NULL, NULL);
		SDL_RenderPresent(sdlRen);
		
		SDL_Event e;
		while(SDL_PollEvent(&e))
		{
			if(e.type == SDL_QUIT)
				exit(0);
		}
		
		
	}
}

void vdac_init()
{
	printf("Dummy VDAC (SDL2)...\n");	
	if(SDL_CreateThread(vdac_renderthread, "VDAC DMA", NULL) == NULL)
	{
		printf("error making render thread\n");
		exit(-1);
	}
	
	printf("Done.\n");

}

void vdac_fieldwait()
{
	int oldfield = vdac_fieldcount;
	do {
		SDL_Delay(1);
	} while(vdac_fieldcount == oldfield);
}

//End of mockup

#endif //JSYS_PC_MOCKUP switch