//iocon.c
//ATTiny48 IO controller source code for BF533 Game System
//Bryan Topp <betopp@betopp.com> 2016

#include <avr/io.h>
#include <avr/interrupt.h>

//Hardware locations on the ATTiny48
#define BIT_SELECT(value, bit) (value & (1<<bit))

#define PIN_P1_LEFT	BIT_SELECT(last_port_d,1) //PD1
#define PIN_P1_RIGHT	BIT_SELECT(last_port_d,3) //PD3
#define PIN_P1_UP	BIT_SELECT(last_port_c,3) //PC3
#define PIN_P1_DOWN	BIT_SELECT(last_port_c,5) //PC5
#define PIN_P1_B1	BIT_SELECT(last_port_a,2) //PA2
#define PIN_P1_B2	BIT_SELECT(last_port_b,6) //PB6
#define PIN_P1_B3	BIT_SELECT(last_port_d,5) //PD5
#define PIN_P1_START	BIT_SELECT(last_port_c,1) //PC1

#define PIN_P2_LEFT	BIT_SELECT(last_port_d,2) //PD2
#define PIN_P2_RIGHT	BIT_SELECT(last_port_d,4) //PD4
#define PIN_P2_UP	BIT_SELECT(last_port_c,4) //PC4
#define PIN_P2_DOWN	BIT_SELECT(last_port_d,0) //PD0
#define PIN_P2_B1	BIT_SELECT(last_port_a,3) //PA3
#define PIN_P2_B2	BIT_SELECT(last_port_b,7) //PB7
#define PIN_P2_B3	BIT_SELECT(last_port_d,6) //PD6
#define PIN_P2_START	BIT_SELECT(last_port_c,2) //PC2

#define PIN_COIN1	BIT_SELECT(last_port_a,1) //PA1
#define PIN_COIN2	BIT_SELECT(last_port_c,0) //PC0
#define PIN_SERVICE	BIT_SELECT(last_port_c,7) //PC7
#define PIN_TEST 	BIT_SELECT(last_port_a,0) //PA0

//Definitions to match those in pads.h for the BF533 software
#define PADS_BIT_LEFT 0x01
#define PADS_BIT_RIGHT 0x02
#define PADS_BIT_UP 0x04
#define PADS_BIT_DOWN 0x08
#define PADS_BIT_B1 0x10
#define PADS_BIT_B2 0x20
#define PADS_BIT_B3 0x40
#define PADS_BIT_START 0x80

#define PADS_SYS_BIT_COIN1 0x01
#define PADS_SYS_BIT_COIN2 0x02
#define PADS_SYS_BIT_SERVICE 0x04
#define PADS_SYS_BIT_TEST 0x08

//Resets a certain bit if a condition is true, sets it if the condition is false.
//Used to reset an output bit when a switch is open (i.e. high) and set it when closed (i.e. low)
#define COND_RESET_SET(condition, output, bitmask) do{if(condition){output &= ~bitmask;}else{output |= bitmask;}}while(0)

//Data available for SPI to request
unsigned char tx_table[16]; //Indexed by low 4 bits of incoming SPI data

int main(void)
{
	//Turn on the internal pull-up resistors for everything
	MCUCR &= ~PUD; //No pull-up disable
	DDRA = 0; //All input, all pins
	DDRB = 0;
	DDRC = 0;
	DDRD = 0;
	PORTA = ~0; //All pull-up, all pins
	PORTB = ~0;
	PORTC = ~0;
	PORTD = ~0;
	
	//Set up the SPI bus, slave mode
	PRR &= ~PRSPI; //Don't power-reduce SPI
	DDRB = (1<<PB4); //Only MISO pin is an output
	SPCR = (1<<SPIE) | (1<<SPE) | (1<<CPHA) | (1<<CPOL); //Enable SPI, enable SPI interrupt, CPOL=CPHA=1
	
	//Dummy reads to clear interrupt flags
	unsigned char dummy = SPDR;
	unsigned char dummy2 = SPSR;
	
	//Storage for IOCON coinage accounting
	unsigned char coincooldown1 = 0; //Debounce counter for first slot
	unsigned char coincooldown2 = 0; //Debounce counter for second slot
	tx_table[0x3] = 0; //Coins accumulated from first slot
	tx_table[0x4] = 0; //Coins accumulated from second slot
	
	//Enable interrupts
	sei(); 
	
	//Main loop - continually read inputs and store them into our memory
	while(1)
	{	
		//Grab inputs		
		unsigned char last_port_a = PINA;
		unsigned char last_port_b = PINB;
		unsigned char last_port_c = PINC;
		unsigned char last_port_d = PIND;

		//Toggle latched bits based on them - decode to a nicer bit ordering
		unsigned char pad1;
		unsigned char pad2;
		unsigned char sysb;
		
		COND_RESET_SET(PIN_P1_LEFT, pad1, PADS_BIT_LEFT); //Pad 1
		COND_RESET_SET(PIN_P1_RIGHT, pad1, PADS_BIT_RIGHT);
		COND_RESET_SET(PIN_P1_UP, pad1, PADS_BIT_UP);
		COND_RESET_SET(PIN_P1_DOWN, pad1, PADS_BIT_DOWN);
		COND_RESET_SET(PIN_P1_B1, pad1, PADS_BIT_B1);
		COND_RESET_SET(PIN_P1_B2, pad1, PADS_BIT_B2);
		COND_RESET_SET(PIN_P1_B3, pad1, PADS_BIT_B3);
		COND_RESET_SET(PIN_P1_START, pad1, PADS_BIT_START);
		
		COND_RESET_SET(PIN_P2_LEFT, pad2, PADS_BIT_LEFT); //Pad 2
		COND_RESET_SET(PIN_P2_RIGHT, pad2, PADS_BIT_RIGHT);
		COND_RESET_SET(PIN_P2_UP, pad2, PADS_BIT_UP);
		COND_RESET_SET(PIN_P2_DOWN, pad2, PADS_BIT_DOWN);
		COND_RESET_SET(PIN_P2_B1, pad2, PADS_BIT_B1);
		COND_RESET_SET(PIN_P2_B2, pad2, PADS_BIT_B2);
		COND_RESET_SET(PIN_P2_B3, pad2, PADS_BIT_B3);
		COND_RESET_SET(PIN_P2_START, pad2, PADS_BIT_START);
		
		COND_RESET_SET(PIN_COIN1, sysb, PADS_SYS_BIT_COIN1); //System
		COND_RESET_SET(PIN_COIN2, sysb, PADS_SYS_BIT_COIN2);
		COND_RESET_SET(PIN_SERVICE, sysb, PADS_SYS_BIT_SERVICE);
		COND_RESET_SET(PIN_TEST, sysb, PADS_SYS_BIT_TEST);
		
		//Ready data for transmitting
		tx_table[0x0] = sysb; //System buttons
		tx_table[0x1] = pad1; //Pad 1
		tx_table[0x2] = pad2; //Pad 2
		//tx_table[0x3] = coincount1; //Coinage Status 1 (Persists through main loop iterations)
		//tx_table[0x4] = coincount2; //Coinage Status 2 (Persists through main loop iterations)
		tx_table[0x5] = 0xAA; //Magic number 1
		tx_table[0x6] = 0x77; //Reserved
		tx_table[0x7] = 0x77; //Reserved
		tx_table[0x8] = 0x77; //Reserved
		tx_table[0x9] = 0x77; //Reserved
		tx_table[0xA] = 0x55; //Magic number 2
		tx_table[0xB] = 0x77; //Reserved
		tx_table[0xC] = 0x77; //Reserved
		tx_table[0xD] = 0x77; //Reserved
		tx_table[0xE] = 0xC0; //Version		
		tx_table[0xF] = 0x77; //Reserved
		
		//Handle coinage		
		if( (!coincooldown1) && (!PIN_COIN1) ) //Cooldown 1 finished, COIN1 switch closed (low) - COIN!
		{
			tx_table[0x3]++; //Coinage status 1
		}
		
		if( (!coincooldown2) && (!PIN_COIN2) ) //Cooldown 2 finished, COIN2 switch closed (low) - COIN!
		{
			tx_table[0x4]++; //Coinage status 2
		}
		
		if(!PIN_COIN1) //Coin1 closed (low), reset cooldown 1
			coincooldown1 = 255;
		else if(coincooldown1 > 0) //Coin1 open (high), decrement cooldown 1 if necessary
			coincooldown1--;
		
		if(!PIN_COIN2) //Coin2 closed (low), reset cooldown 2
			coincooldown2 = 255;
		else if(coincooldown2 > 0) //Coin2 open (high), decrement cooldown 2 if necessary
			coincooldown2--;
		
		
		
	}

	return 0;
}

//Absolutely minimal, for the sake of the 400MHz Blackfin core waiting on us
ISR(SPI_STC_vect)
{
	//Respond based on byte in
	unsigned char command = SPDR % 16;
	SPDR = tx_table[command];
	
	//Smash over the entry in the command table, atomically with it being sent over SPI.
	//Accumulated coinage is thus guaranteed to be sent exactly once.
	tx_table[command] = 0; 
}
