//sbl533.c
//Simple bootloader for BF533 development
//Initializes processor and UART for bootstrapping
//Bryan Topp <betopp@betopp.com> 2016

#ifndef __ADSPLPBLACKFIN__
	#error This is for the BF533
#endif

#include <sys/platform.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#define UART_BAUD 115200 //Desired UART baud rate

#define CMDBUF_LEN 256
char cmdbuf[CMDBUF_LEN] = {0};
unsigned short cmdbuf_pos = 0;

char *cmdbuf_tok0 = NULL;
char *cmdbuf_tok1 = NULL;
char *cmdbuf_tok2 = NULL;

//Crappy flow control
#define XOFF_C 0x13
#define XON_C 0x11


//UART driving functions
//----------------------

//Waits until we can transmit a byte
static void uart_wait_txrdy()
{
	//Wait for the transmit buffer to be nonfull
	while (!(*pUART_LSR & THRE)) {};
}

//Waits until we can receive a byte
static void uart_wait_rxrdy()
{
	//Wait for the receive buffer to be nonempty
	while (!(*pUART_LSR & DR)) {};
}

//Transmits a byte (and hacks around line-ending differences - I like to write '\n')
static void uart_putc(char c)
{
	uart_wait_txrdy();
	//Place one character in the transmit buffer
	*pUART_THR = c;
	
	//Hack
	if(c == '\n')
		uart_putc('\r');
}

//Receives a byte
static char uart_getc()
{
	uart_wait_rxrdy();
	//Grab one character from the receive buffer
	return *pUART_RBR;
}

//String input/output functions - printf() and friends are way too heavyweight
//----------------------------------------------------------------------------

//Outputs a NUL-terminated string to the UART
static void uart_puts(char *str)
{
	for(int i = 0; str[i] != '\0'; i++)
	{
		uart_putc(str[i]);
	}
}

//Like uart_puts, but with a maximum length to output
static void uart_putsn(char *str, int n)
{
	for(int i = 0; i < n && str[i] != '\0'; i++)
	{
		uart_putc(str[i]);
	}
}

//Prints a 32-bit hex number
void printhex32(unsigned int n)
{	
	uart_putc('0');
	uart_putc('x');
	
	for(int d = 0; d < 8; d++)
	{
		unsigned char h = n >> 28;
		if(h <= 0x9)
			uart_putc('0' + h);
		else
			uart_putc('A' - 10 + h);
		
		n = n << 4;
	}
}

//Prints a 16-bit hex number
void printhex16(unsigned short n)
{	
	uart_putc('0');
	uart_putc('x');
	
	for(int d = 0; d < 4; d++)
	{
		unsigned char h = n >> 12;
		if(h <= 0x9)
			uart_putc('0' + h);
		else
			uart_putc('A' - 10 + h);
		
		n = n << 4;
	}
}

//Prints a 16-bit binary number
void printbin16(unsigned int n)
{
	uart_putc('%');
	
	for(int b = 15; b >= 0; b--)
	{
		if(n & (1<<b))
			uart_putc('1');
		else
			uart_putc('0');
		
		if(b % 4 == 0 && b != 0)
			uart_putc(' ');
	}
}

//Prints a 32-bit binary number
void printbin32(unsigned int n)
{
	uart_putc('%');
	
	for(int b = 31; b >= 0; b--)
	{
		if(n & (1<<b))
			uart_putc('1');
		else
			uart_putc('0');
		
		if(b % 4 == 0 && b != 0)
			uart_putc(' ');
	}
}

//Prints a 32-bit number as decimal
void printdec(unsigned int n)
{
	char buf[16]; //eh
	int d = 0;
	
	if(n == 0)
	{
		uart_putc('0');
	}
	else
	{
		while(n > 0)
		{
			buf[d] = (n % 10) + '0';
			d++;
			n = n / 10;
		}
		
		while(d > 0)
		{
			d--;
			uart_putc(buf[d]);
		}
	}
}

//Prints an 8-bit hex number
void printhex8(unsigned char n)
{
	uart_putc('0');
	uart_putc('x');

	unsigned char h = n >> 4;
	if(h <= 0x9)
		uart_putc('0' + h);
	else
		uart_putc('A' - 10 + h);
	
	unsigned char l = n & 0xF;
	if(l <= 0x9)
		uart_putc('0' + l);
	else
		uart_putc('A' - 10 + l);
}

//Prints an 8-bit binary number
void printbin8(unsigned char n)
{
	uart_putc('%');
	
	for(int b = 7; b >= 0; b--)
	{
		if(n & (1<<b))
			uart_putc('1');
		else
			uart_putc('0');
		
		if(b % 4 == 0 && b != 0)
			uart_putc(' ');
	}
}

//Tries to be atoi(), but with base16/base2 support
unsigned int readval(const char *str)
{
	if(str[0] == '0' && (str[1] == 'x' || str[1] == 'X') ) //Hex
		return strtoul(str+2, NULL, 16);
	else if(str[0] == '%') //Binary
		return strtoul(str+1, NULL, 2);
	else
		return strtoul(str, NULL, 10);
}

//Command helpers - called from command parsing loop
//--------------------------------------------------

//PEEK word (32 bits)
void peekw()
{
	unsigned int *p = (unsigned int*)readval(cmdbuf_tok1);
	if((unsigned int)p % 4)
	{
		uart_puts("Error - word reads must be naturally-aligned.\n");
		return;
	}
	
	unsigned int v = *p;
	printhex32((unsigned int)p);
	uart_puts(": ");
	printhex32(v);
	uart_puts(" / ");
	printbin32(v);
	uart_puts(" / ");
	printdec(v);
	uart_puts("\n");
}

//POKE word (32 bits)
void pokew()
{
	unsigned int *p = (unsigned int*)readval(cmdbuf_tok1);
	if((unsigned int)p % 4)
	{
		uart_puts("Error - word writes must be naturally-aligned.\n");
		return;
	}
	
	
	unsigned int v = (unsigned int)readval(cmdbuf_tok2);
	*p = v;
	peekw();	
}

//PEEK halfword (16 bits)
void peekh()
{
	unsigned short *p = (unsigned short*)readval(cmdbuf_tok1);
	if((unsigned int)p % 2)
	{
		uart_puts("Error - halfword reads must be naturally-aligned.\n");
		return;
	}
	
	unsigned short v = *p;
	printhex16((unsigned int)p);
	uart_puts(": ");
	printhex16(v);
	uart_puts(" / ");
	printbin16(v);
	uart_puts(" / ");
	printdec(v);
	uart_puts("\n");
}

//POKE halfword (16 bits)
void pokeh()
{
	unsigned short *p = (unsigned short*)readval(cmdbuf_tok1);
	if((unsigned int)p % 2)
	{
		uart_puts("Error - halfword writes must be naturally-aligned.\n");
		return;
	}
	
	
	unsigned short v = (unsigned short)readval(cmdbuf_tok2);
	*p = v;
	peekh();	
}

//PEEK byte (8 bits)
void peekb()
{
	unsigned char *p = (unsigned char*)readval(cmdbuf_tok1);
	unsigned char v = *p;
	printhex32((unsigned int)p);
	uart_puts(": ");
	printhex8(v);
	uart_puts(" / ");
	printbin8(v);
	uart_puts(" / ");
	printdec(v);
	uart_puts("\n");
}

//POKE byte (8 bits)
void pokeb()
{
	unsigned char *p = (unsigned char*)readval(cmdbuf_tok1);
	unsigned char v = (unsigned char)readval(cmdbuf_tok2);
	*p = v;
	peekb();
}


//DRAM initialization procedure
//
void draminit()
{		
	//Disable all asynchronous banks
	*pEBIU_AMGCTL = 0;
	
	//Clear SDRAM status register sticky bits
	*pEBIU_SDSTAT = 0xFFFF;
	
	//Refresh rate control [BF533 HW Ref p. 17-48]
	//F_SCLK = 98.304MHz
	//T_REF = 64ms
	//NRA = 8192
	//T_RAS = 5
	//T_RP = 3
	//RDIV = ((F_SCLK x T_REF ) / NRA) - (T_RAS + T_RP )
	//0xFFC00A18 0x2F8
	*pEBIU_SDRRC = 760;
	
	//Set up bank - 4M x 16 x 4 banks - AS4C16M16SA
	//0xFFC00A14 0x13
	*pEBIU_SDBCTL = EBE | EBSZ_32 | EBCAW_9; 
	
	//Timings for AS4C16M16SA-7 at 98.304MHz SCLK / power-on sequence
	//0xFFC00A10 = 0x919949 / reads back as 0x119949
	*pEBIU_SDGCTL = SCTLE | CL_2 | TRAS_5 | TRP_3 | TRCD_3 | TWR_2 | PSS;
	//GOD DAMNIT, it's PSS, not "PSSE!" That's overloaded elsewhere.
	
	ssync();
	
	//Trigger a read from SDRAM to start it
	unsigned short a = *(volatile unsigned short*)0; //BREAKIN' THE LAW BREAKIN' THE LAW
	a = a; //Suppress compiler warning
}

unsigned int lfsr32_advance(unsigned int old)
{
	if(old & 1)
	{
		return (old>>1) ^ (0xA3000000 /*Galois taps*/);
	}
	else
	{
		return old>>1;
	}
	
}

unsigned int lfsr24_advance(unsigned int old)
{
	old &= 0x00FFFFFF;
	
	if(old & 1)
	{
		return (old>>1) ^ (0x00D80000 /*Galois taps*/);
	}
	else
	{
		return old>>1;
	}
	
}

void dram_print_error(volatile unsigned short *address, unsigned short existing_data, unsigned short expected_data)
{
	uart_puts("Error! RD: ");
	printhex32((unsigned int)address);
	uart_puts(" = ");
	printhex16(existing_data);
	
	uart_puts("(EX: ");
	printhex16(expected_data);
	uart_puts(")\n");
}

//DRAM testing function
void dramtest()
{
	volatile unsigned short *dram = 0;
	
	uart_puts("Writing LFSR pattern...\n");
	
	unsigned int addr_prng = 1234u;
	unsigned int data_prng = 4321u;
		
	//Write pseudorandom data at pseudorandom addresses
	do {		
		dram[addr_prng] = data_prng;
		
		addr_prng = lfsr24_advance(addr_prng);
		data_prng = lfsr32_advance(data_prng);
	} while(addr_prng != 1234u);
	
	//Reset the PRNGs
	addr_prng = 1234u;
	data_prng = 4321u;
	
	uart_puts("Writes finished, reading and verifying...\n");
	
	//Read and verify
	do {	
		unsigned short existing_data = dram[addr_prng];
		unsigned short expected_data = (unsigned short)data_prng;
		
		if(existing_data != expected_data)
		{
			dram_print_error(&(dram[addr_prng]), existing_data, expected_data);
			return;
		}
		
		addr_prng = lfsr24_advance(addr_prng);
		data_prng = lfsr32_advance(data_prng);
	} while(addr_prng != 1234u);
	
	uart_puts("Success.\n");
}


//SPI slave select pin locations
#define SPI_SLAVE_ROM_ENABLE FLS2 //SPI_ROMSEL = PF2
#define SPI_SLAVE_FRAM_ENABLE FLS1 //SPI_FRAMSEL = PF1
#define SPI_SLAVE_IOCON_ENABLE FLS3 //SPI_IOSEL = PF3
#define SPI_SLAVE_ENABLES (SPI_SLAVE_ROM_ENABLE | SPI_SLAVE_FRAM_ENABLE | SPI_SLAVE_IOCON_ENABLE)

//Active-low chip selects
#define SPI_SLAVE_ROM_SELECT ((0xFF00 & ~FLG2) | SPI_SLAVE_ENABLES)
#define SPI_SLAVE_FRAM_SELECT ((0xFF00 & ~FLG1) | SPI_SLAVE_ENABLES)
#define SPI_SLAVE_IOCON_SELECT ((0xFF00 & ~FLG3) | SPI_SLAVE_ENABLES)
#define SPI_SLAVE_NONE_SELECT (0xFF00 | SPI_SLAVE_ENABLES)

//IORST = PF0

void spiinit()
{
	//Set slave-select pins active and deselect all
	*pSPI_FLG = SPI_SLAVE_NONE_SELECT;
	
	//System clock is 98.304MHz
	//Set SPI_BAUD to 32 -> division by 64 -> 1.536MHz SPI baud
	//(Baud rate under ~2MHz necessary for IOCON - maybe push higher for flash loading?)
	*pSPI_BAUD = 32;

	//SPI Master mode
	//Enable SPI
	//SPI clock phase 1
	//SPI clock polarity 1
	//Transmit zeroes when transmit buffer empties
	//Initiate on transmit buffer write
	*pSPI_CTL = MSTR | SPE | CPHA | CPOL | SZ | TDBR_CORE;
}

void spiselect()
{
	if(cmdbuf_tok1 == NULL || cmdbuf_tok1[0] == 'n')
	{
		uart_puts("Deselect...\n");
		*pSPI_FLG = SPI_SLAVE_NONE_SELECT;
	}
	else if(cmdbuf_tok1[0] == 'r')
	{
		uart_puts("ROM...\n");
		*pSPI_FLG = SPI_SLAVE_ROM_SELECT;
	}
	else if(cmdbuf_tok1[0] == 'f')
	{
		uart_puts("FRAM...\n");
		*pSPI_FLG = SPI_SLAVE_FRAM_SELECT;
	}
	else if(cmdbuf_tok1[0] == 'i')
	{
		uart_puts("IOCON...\n");
		*pSPI_FLG = SPI_SLAVE_IOCON_SELECT;
	}
	else
	{
		uart_puts("Unknown device specified, not changing.\n");
		return;
	}
}

void spiw()
{
	
	//Decide what we're going to transmit
	unsigned char out_byte = 0;
	if(cmdbuf_tok1 != NULL)
	{
		out_byte = readval(cmdbuf_tok1);
	}

	printhex8(out_byte);
	uart_puts("...");
	
	//Transmit it
	*pSPI_STAT &= ~RXS;
	*pSPI_TDBR = out_byte;
	
	//Wait
	ssync();
	while(!(*pSPI_STAT & RXS)) {}
	ssync();
		
	//Get what came back
	unsigned char in_byte = *pSPI_RDBR;
		
	printhex8(in_byte);
	uart_puts("\n");
}


void iorstb()
{
	if(cmdbuf_tok1 == NULL)
	{
		uart_puts("No value specified.\n");
		return;
	}
	
	
	*pFIO_DIR = ~0; //All flags output
	
	unsigned char val = readval(cmdbuf_tok1);
	
	if(val)
	{
		uart_puts("Setting PF0 - IOCON reset deasserted/high\n");
		*pFIO_FLAG_S = PF0;
	}
	else
	{
		uart_puts("Clearing PF0 - IOCON reset asserted/low\n");
		*pFIO_FLAG_C = PF0;
	}

}



unsigned char spi_exchange(unsigned char tx)
{	
	*pSPI_STAT &= ~RXS;
	*pSPI_TDBR = tx;
	
	//Wait
	ssync();
	while(!(*pSPI_STAT & RXS)) {}
	ssync();
		
	//Get what came back
	return *pSPI_RDBR;
}

void iocon_probe()
{
	//SPI setup: not using any SPI chip-selects right now
	*pSPI_FLG = 0xFF00;
	*pSPI_BAUD = 256; //Go real slow
	*pSPI_CTL = MSTR | SPE | SZ | TDBR_CORE; //CPHA=CPOL=0
	
	*pFIO_DIR = ~0; //All flags output	

	*pFIO_FLAG_D = 0xFFFF; //All flags high
	ssync();
	for(int nop = 0; nop < 3500000; nop++){	asm("nop;"); }
	
	*pFIO_FLAG_D = 0xFFFE; //PF0 low, putting IOCON in reset
	ssync();
	for(int nop = 0; nop < 3500000; nop++){	asm("nop;"); }

	//Send programming sequence
	unsigned char a = spi_exchange(0xAC);	
	unsigned char b = spi_exchange(0x53);
	unsigned char c = spi_exchange(0x00);
	unsigned char d = spi_exchange(0x00);
	
	printhex8(a); uart_puts("\n");
	printhex8(b); uart_puts("\n");
	printhex8(c); uart_puts("\n");
	printhex8(d); uart_puts("\n");
	
}

//Kermit downloader
void kdl()
{
	unsigned int dest_addr;
	
	if(cmdbuf_tok1 == NULL)
	{
		uart_puts("No address given. Defaulting to 0x01000000.\n");
		dest_addr = 0x01000000;
	}
	else
	{
		dest_addr = readval(cmdbuf_tok1);
	}
	
	unsigned int *pDest = (unsigned int*)dest_addr;
	
	//Make a prompt that looks like the uBoot one
	uart_puts("## Ready for binary (kermit) download to ");
	printhex32(dest_addr);
	uart_puts(" at 115200 bps...\n");

	#define KSTATE_R 0
	#define KSTATE_RF 1
	#define KSTATE_RD 2
	#define KSTATE_C 3
	#define KSTATE_A 4
	#define KSTATE_EXIT 0xFF
	
	
	unsigned char kermit_state = 0;

	while(kermit_state != KSTATE_EXIT) //Kermit loop
	{		
		//Grab a packet...
		
		//Wait for the MARK
		uart_putc(XON_C);
		while(1)
		{
			unsigned char mark = uart_getc();
			if(mark == 0x01)
				break;
		}
		
		//Get the length - and start running the checksum
		unsigned char len_plus_32 = uart_getc();
		unsigned char checksum = len_plus_32;

		//Get the sequence number and checksum it
		unsigned char seq_plus_32 = uart_getc();
		checksum += seq_plus_32;
		
		//Get the type and checksum it
		unsigned char type = uart_getc();
		checksum += type;
		
		//Get all the data
		unsigned char data[96];
		unsigned char data_pos = 0;
		unsigned char data_len = (len_plus_32 - 32) - 3;
		while(data_pos < data_len)
		{
			data[data_pos] = uart_getc();
			checksum += data[data_pos];
			
			if(data[data_pos] < 32) //Control chars not allowed in data
				break;
			
			data_pos++;
		}
		
		//Get the check byte
		unsigned char check = uart_getc();
		unsigned char check_computed = ((checksum + ((checksum & 192)/64)) & 63) + 32;
		
		if(check_computed != check)
		{
			uart_puts("Bad!\n");
		}
		
		//Process it
		uart_putc(XOFF_C);
		switch(kermit_state)
		{
			case KSTATE_R:
			{
			
				break;
			}
			case KSTATE_RF:
			{
			
				break;
			}
			case KSTATE_RD:
			{
				
				break;
			}
			case KSTATE_C:
			{
				kermit_state = KSTATE_EXIT;
				break;
			}
			case KSTATE_A:
			{
				kermit_state = KSTATE_EXIT;
				break;
			}
		}
	}
}



//Main program follows - initialization, loop, and command parsing
//----------------------------------------------------------------

void runcmd()
{
	//Tokenize the command buffer
	cmdbuf_tok0 = strtok(cmdbuf, " ");
	cmdbuf_tok1 = strtok(NULL, " ");
	cmdbuf_tok2 = strtok(NULL, " ");
	
	if(cmdbuf_tok0 == NULL || cmdbuf_tok0[0] == '\0')
		return;
	
	if(!strncmp(cmdbuf_tok0, "help", CMDBUF_LEN))
	{
		uart_puts("Available commands:\n");
		uart_puts("help             - Show this information\n");
		uart_puts("reset            - Reboot the system\n");
		uart_puts("version          - Show build information.\n");
		uart_puts("peekb $A         - Show byte at given memory address\n");
		uart_puts("pokeb $A $V      - Write a byte to the given address\n");
		uart_puts("peekh $A         - Peek (as above) a halfword (16 bits)\n");
		uart_puts("pokeh $A $V      - Poke (as above) a halfword (16 bits)\n");
		uart_puts("peekw $A         - Peek (as above) a word (32 bits)\n");
		uart_puts("pokew $A $V      - Poke (as above) a word (32 bits)\n");
		uart_puts("draminit         - Set up on-board DRAM\n");
		uart_puts("dramtest         - Run test of on-board DRAM, clobbering its contents\n");
		uart_puts("dramstat         - Show DRAM status register\n");
		uart_puts("spiinit          - Set up SPI controller\n");
		uart_puts("spiselect [rfin] - Select SPI slave ROM/FRAM/IOCON/none\n");
		uart_puts("spiw $B          - Write a byte to SPI and show the result\n");
		uart_puts("iorstb $V        - Set value of ~IORST flag (0 to reset, 1 to run)\n");
		uart_puts("iocon_probe      - Probe IO controller programming interface\n");
		uart_puts("kdl $A           - Start a Kermit download to the specified address\n");
	}
	else if(!strncmp(cmdbuf_tok0, "reset", CMDBUF_LEN))
	{
		uart_puts("Resetting...\n");
		*pSWRST = 7;
		ssync();
		for(short i = 0; i < 100; i++)
			asm("nop;");
			
		*pSWRST = 0;
		ssync();
		asm("raise 1;");
	}
	else if(!strncmp(cmdbuf_tok0, "version", CMDBUF_LEN))
	{
		uart_puts("-----------------------------------------------\n");
		uart_puts("SBL533: Simple bootloader for BF533 Game System\n");
		uart_puts("-----------------------------------------------\n");
		uart_puts("Version: " SBL_GIT_VERSION "\n");
		uart_puts("Built:   " SBL_BUILD_DATE "\n");
		uart_puts("Builder: " SBL_BUILDER "\n");
		uart_puts("(c)2016 Bryan E. Topp / Topp Technical Products\n");
		uart_puts("-----------------------------------------------\n");
	}
	else if(!strncmp(cmdbuf_tok0, "peekb", CMDBUF_LEN))
	{
		peekb();
	}
	else if(!strncmp(cmdbuf_tok0, "peekw", CMDBUF_LEN))
	{
		peekw();
	}
	else if(!strncmp(cmdbuf_tok0, "peekh", CMDBUF_LEN))
	{
		peekh();
	}
	else if(!strncmp(cmdbuf_tok0, "pokeb", CMDBUF_LEN))
	{
		pokeb();		
	}
	else if(!strncmp(cmdbuf_tok0, "pokew", CMDBUF_LEN))
	{
		pokew();
	}
	else if(!strncmp(cmdbuf_tok0, "pokeh", CMDBUF_LEN))
	{
		pokeh();
	}
	else if(!strncmp(cmdbuf_tok0, "draminit", CMDBUF_LEN))
	{
		draminit();
	}
	else if(!strncmp(cmdbuf_tok0, "dramtest", CMDBUF_LEN))
	{
		dramtest();
	}
	else if(!strncmp(cmdbuf_tok0, "dramstat", CMDBUF_LEN))
	{
		printhex32(*pEBIU_SDSTAT);
		uart_puts(" / ");
		printbin32(*pEBIU_SDSTAT);
		uart_puts("\n");
		
		uart_puts("Writing ~0 to status register...\n");
		*pEBIU_SDSTAT = 0xFFFF;
		
		printhex32(*pEBIU_SDSTAT);
		uart_puts(" / ");
		printbin32(*pEBIU_SDSTAT);
		uart_puts("\n");
		
	}
	else if(!strncmp(cmdbuf_tok0, "spiinit", CMDBUF_LEN))
	{
		spiinit();
	}
	else if(!strncmp(cmdbuf_tok0, "spiw", CMDBUF_LEN))
	{
		spiw();
	}
	else if(!strncmp(cmdbuf_tok0, "spiselect", CMDBUF_LEN))
	{
		spiselect();
	}
	else if(!strncmp(cmdbuf_tok0, "iorstb", CMDBUF_LEN))
	{
		iorstb();
	}
	else if(!strncmp(cmdbuf_tok0, "iocon_probe", CMDBUF_LEN))
	{
		iocon_probe();
	}
	else if(!strncmp(cmdbuf_tok0, "kdl", CMDBUF_LEN))
	{
		kdl();
	}
	else
	{
		uart_puts("Unknown command: ");
		uart_putsn(cmdbuf_tok0, CMDBUF_LEN);
		uart_puts("\n");
	}
}

int main(int argc, const char **argv)
{
	//Comments for hardware init section can be found in the main project.
	
	//Init PLL/Dividers
	*pPLL_CTL = SET_MSEL(32);
	asm("CLI R0 ;");
	asm("IDLE ;");
	asm("STI R0 ;");
	*pPLL_DIV = SET_SSEL(4);
	
	//Init UART
	*pUART_GCTL = UCEN;
	
	unsigned short divisor = (unsigned short)(98304000u/(16u*UART_BAUD) + 1u);
	*pUART_LCR = DLAB;
	*pUART_DLL = divisor & 0xFF;    //Divisor Low
	*pUART_DLH = (divisor>>8) & 0xFF; //Divisor High
	
	*pUART_LCR = WLS(8);
	
	//Print intro
	uart_putc(XOFF_C);
	uart_puts("Loaded SBL533.\n");
	
	//Wait for commands	
	while(1) //Outer loop - each command
	{
		//Clear command buffer
		memset(cmdbuf, '\0', CMDBUF_LEN);
		cmdbuf_pos = 0;
		
		//Show prompt
		uart_puts("SBL533 # ");
		uart_putc(XON_C);
		
		//Inner loop - editing the command
		while(1)
		{
			//Grab character from input
			unsigned char nextchar = uart_getc();
			
			if(nextchar == '\r') //EOL, execute command
			{
				uart_puts("\n");
				break;
			}
			else if(nextchar == '\b') //Backspace
			{
				if(cmdbuf_pos > 0) //Don't allow backspacing too far
				{
					cmdbuf[cmdbuf_pos-1] = '\0'; //Erase last character
					cmdbuf_pos--;
					
					uart_putc('\b'); //Move back
					uart_putc(' '); //White-out
					uart_putc('\b'); //Back up again
				}
			}
			else if(!iscntrl(nextchar)) //Normal character that should be appended to the command buffer
			{
				if(cmdbuf_pos < CMDBUF_LEN-1) //Don't allow us to run off the end; leave room for \0
				{
					cmdbuf[cmdbuf_pos] = nextchar;
					cmdbuf_pos++;
					
					uart_putc(nextchar); //Echo the character back to show that it was accepted
				}				
			}
			else
			{
				//Ignore other control characters
			}
		}
		
		//Run command
		uart_putc(XOFF_C);
		runcmd();
	}
}